import threading
import logging
import json
import numpy as np
import time
import dbus
import dbus.service
import dbus.mainloop.glib
from gi.repository import GLib
from ml_tools.tools import CustomJSONEncoder
from cptv import Frame

from dbus.mainloop.glib import DBusGMainLoop

DBUS_NAME = "org.cacophony.thermalrecorder"
DBUS_PATH = "/org/cacophony/thermalrecorder"


class Service(dbus.service.Object):
    def __init__(
        self, dbus, get_frame, headers, take_snapshot_fn, labels, get_thumbnail
    ):
        super().__init__(dbus, DBUS_PATH)
        self.get_frame = get_frame
        self.get_thumbnail = get_thumbnail
        self.headers = headers
        self.take_snapshot = take_snapshot_fn
        self.labels = labels

    @dbus.service.method(
        DBUS_NAME,
        in_signature="",
        out_signature="a{si}",
    )
    def CameraInfo(self):
        logging.debug("Serving headers %s", self.headers)
        headers = self.headers.as_dict()
        ir = headers.get("model") == "IR"
        for k, v in headers.items():
            try:
                headers[k] = int(v)
            except:
                headers[k] = 0
                pass
        headers["FPS"] = headers.get("fps", 9)
        headers["ResX"] = headers.get("res_x", 160)
        headers["ResY"] = headers.get("res_y", 120)
        if ir:
            headers["Model"] = 2
        else:
            headers["Model"] = 1
        logging.debug("Sending headers %s", headers)
        return headers

    @dbus.service.method(
        DBUS_NAME,
        in_signature="i",
        out_signature="(aaq(xsiqddxb)s)",
    )
    def TakeSnapshot(self, last_num):
        s = time.time()
        last_frame, track_meta, f_num = self.get_frame(last_num)

        if f_num == last_num or last_frame is None:
            return (np.empty((0, 0)), (0, "", f_num, 0, 0, 0, 0, False), "")
        logging.debug(
            "Frame requested %s latest frame %s took %s",
            last_num,
            f_num,
            time.time() - s,
        )
        if not isinstance(last_frame, Frame):
            last_frame = last_frame[:, :, 0]
            return (
                last_frame,
                (0, "", f_num, 0, 0, 0, 0, 0),  # count
                json.dumps(track_meta, cls=CustomJSONEncoder),
            )
        return (
            last_frame.pix,
            (
                last_frame.time_on.total_seconds() * 1e9,
                "",
                f_num,  # count
                0,
                last_frame.temp_c,
                last_frame.last_ffc_temp_c,
                last_frame.last_ffc_time.total_seconds() * 1e9,
                last_frame.background_frame,
            ),
            json.dumps(track_meta, cls=CustomJSONEncoder),
        )

    @dbus.service.method(
        DBUS_NAME,
        out_signature="aaqiai",
    )
    def GetThumbnail(self, clip_id, track_id):
        if track_id == 0:
            track_id = None
        if clip_id == 0:
            clip_id = None
        result = self.get_thumbnail(clip_id, track_id)
        if result is None:
            raise Exception("No thumbnail")
        thumb, track_id, region = result

        return thumb, track_id, region.to_ltrb()

    @dbus.service.method(
        DBUS_NAME,
    )
    def TakeTestRecording(self):
        logging.info("Take test recording")
        result = False
        try:
            result = self.take_snapshot()
        except:
            logging.error("Error taking test recording", exc_info=True)

        return result

    @dbus.service.method(DBUS_NAME, signature="a{ias}")
    def ClassificationLabels(self):
        logging.info("Getting labels %s", self.labels)
        return self.labels

    @dbus.service.signal(DBUS_NAME, signature="iiaisiaiiibbii")
    def Tracking(
        self,
        clip_id,
        track_id,
        prediction,
        what,
        confidence,
        region,
        frame,
        mass,
        blank,
        tracking,
        last_prediction_frame,
        model_id,
    ):
        pass

    @dbus.service.signal(DBUS_NAME, signature="xb")
    def Recording(self, timestamp, is_recording):
        pass


class SnapshotService:
    def __init__(self, get_frame, headers, take_snapshot_fn, labels, get_thumbnail):
        DBusGMainLoop(set_as_default=True)
        dbus.mainloop.glib.threads_init()
        self.loop = GLib.MainLoop()
        self.t = threading.Thread(
            target=self.run_server,
            args=(get_frame, headers, take_snapshot_fn, labels, get_thumbnail),
        )
        self.t.start()
        self.service = None

    def quit(self):
        self.loop.quit()

    def run_server(self, get_frame, headers, take_snapshot_fn, labels, get_thumbnail):
        session_bus = dbus.SystemBus(mainloop=DBusGMainLoop())
        name = dbus.service.BusName(DBUS_NAME, session_bus)
        self.service = Service(
            session_bus, get_frame, headers, take_snapshot_fn, labels, get_thumbnail
        )
        self.loop.run()

    def tracking(
        self,
        clip_id,
        track,
        prediction,
        region,
        tracking,
        last_prediction_frame,
        labels,
        model_id,
    ):
        logging.debug(
            "Tracking?  %s region %s prediction %s track %s",
            tracking,
            region,
            prediction,
            track.get_id(),
        )
        if self.service is None:
            return
        if prediction is not None:
            predictions = prediction.copy()
            predictions = np.uint8(np.round(predictions * 100))
            best = np.argmax(predictions)
            self.service.Tracking(
                clip_id,
                track.get_id(),
                predictions,
                labels[best],
                predictions[best],
                region.to_ltrb(),
                region.frame_number,
                region.mass,
                region.blank,
                tracking,
                last_prediction_frame,
                model_id,
            )
        else:
            self.service.Tracking(
                clip_id,
                track.get_id(),
                [],
                "",
                0,
                region.to_ltrb(),
                region.frame_number,
                region.mass,
                region.blank,
                tracking,
                last_prediction_frame,
                0,
            )

    def recording(self, is_recording):
        if self.service is None:
            return
        self.service.Recording(np.int64(time.time()), is_recording)



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2020, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from abc import ABC, abstractmethod

from .service import SnapshotService


class Processor(ABC):
    def __init__(
        self,
    ):
        model_labels = {}
        if self.classifier is not None:
            model_labels[self.classifier.id] = self.classifier.labels
        if self.fp_model is not None:
            model_labels[self.fp_model.id] = self.fp_model.labels

        self.service = SnapshotService(
            self.get_recent_frame,
            self.headers,
            self.take_snapshot,
            model_labels,
            self.get_thumbnail,
        )

    @abstractmethod
    def take_snapshot(self): ...

    @abstractmethod
    def process_frame(self, lepton_frame): ...

    @abstractmethod
    def get_recent_frame(self, last_frame=None): ...

    @abstractmethod
    def disconnected(self): ...

    @abstractmethod
    def skip_frame(self): ...

    @property
    @abstractmethod
    def res_x(self): ...

    @property
    @abstractmethod
    def res_y(self): ...

    @property
    @abstractmethod
    def output_dir(self): ...



"""
Author: Giampaolo Ferraro
A script which downloads cptv files and metadata and creates tracking tests
automatically from the provided track information. Note if the tracking is not
perfect in the download cptv files the generated tests will need to be manually
altered
"""

import os
import yaml
import argparse

from pathlib import Path
from .testconfig import TestRecording, TestConfig
from cacophonyapi.user import UserAPI

from .trackingtest import iter_to_file


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("out_folder", help="Root folder to place downloaded files in.")
    parser.add_argument("user", help="API server username")
    parser.add_argument("password", help="API server password")
    parser.add_argument(
        "-s",
        "--server",
        default="https://api.cacophony.org.nz",
        help="CPTV file server URL",
    )
    parser.add_argument(
        "-t",
        "--test_file",
        default="tracking-tests.yml",
        help="File to save generated tests to",
    )
    parser.add_argument(
        "ids",
        nargs="+",
        help="List of recording ids to download",
    )

    return parser.parse_args()


def main():
    args = parse_args()
    api = UserAPI(args.server, args.user, args.password)
    out_dir = Path(args.out_folder)
    tests = []
    test_data = TestConfig(
        recording_tests=tests, server=args.server, clip_dir=args.out_folder
    )
    for rec_id in args.ids:
        rec_id = rec_id.strip()
        rec_meta = api.get(rec_id)
        tracks = api.get_tracks(rec_id)
        filename = Path(rec_id).with_suffix(".cptv")
        fullpath = out_dir / filename
        tests.append(TestRecording.load_from_meta(rec_meta, tracks, filename))
        if not fullpath.exists():
            if iter_to_file(fullpath, api.download_raw(rec_id)):
                print("Saved {} - {}".format(rec_id, fullpath))
            else:
                print("error saving {}".format(rec_id))
    if os.path.exists(args.test_file):
        os.remove(args.test_file)
    with open(args.test_file, "w") as f:
        yaml.dump(test_data, f)


if __name__ == "__main__":
    main()



# https://medium.com/analytics-vidhya/understanding-and-implementation-of-residual-networks-resnets-b80f9a507b9c
import tensorflow as tf


def ResNet50(X_input, classes=6):
    """
    Implementation of the popular ResNet50 the following architecture:
    CONV2D -> BATCHNORM -> RELU -> MAXPOOL -> CONVBLOCK -> IDBLOCK*2 -> CONVBLOCK -> IDBLOCK*3
    -> CONVBLOCK -> IDBLOCK*5 -> CONVBLOCK -> IDBLOCK*2 -> AVGPOOL -> TOPLAYER

    Arguments:
    input_shape -- shape of the images of the dataset
    classes -- integer, number of classes

    Returns:
    model -- a Model() instance in Keras
    """

    # Define the input as a tensor with shape input_shape

    # Zero-Padding
    X = tf.keras.layers.ZeroPadding2D((3, 3))(X_input)

    # Stage 1
    X = tf.keras.layers.Conv2D(
        64,
        (7, 7),
        strides=(2, 2),
        name="conv1",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X)
    X = tf.keras.layers.BatchNormalization(axis=3, name="bn_conv1")(X)
    X = tf.keras.layers.Activation("relu")(X)
    X = tf.keras.layers.MaxPooling2D((3, 3), strides=(2, 2))(X)

    # Stage 2
    X = convolutional_block(X, f=3, filters=[64, 64, 256], stage=2, block="a", s=1)
    X = identity_block(X, 3, [64, 64, 256], stage=2, block="b")
    X = identity_block(X, 3, [64, 64, 256], stage=2, block="c")

    # Stage 3
    X = convolutional_block(X, f=3, filters=[128, 128, 512], stage=3, block="a", s=2)
    X = identity_block(X, 3, [128, 128, 512], stage=3, block="b")
    X = identity_block(X, 3, [128, 128, 512], stage=3, block="c")
    X = identity_block(X, 3, [128, 128, 512], stage=3, block="d")

    # Stage 4
    X = convolutional_block(X, f=3, filters=[256, 256, 1024], stage=4, block="a", s=2)
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block="b")
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block="c")
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block="d")
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block="e")
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block="f")

    # Stage 5
    X = convolutional_block(X, f=3, filters=[512, 512, 2048], stage=5, block="a", s=2)
    X = identity_block(X, 3, [512, 512, 2048], stage=5, block="b")
    X = identity_block(X, 3, [512, 512, 2048], stage=5, block="c")

    # AVGPOOL . Use "X = AveragePooling2D(...)(X)"
    X = tf.keras.layers.AveragePooling2D()(X)

    # output layer
    X = tf.keras.layers.Flatten()(X)
    # X = tf.keras.layers.Dense(
    #     classes,
    #     activation="softmax",
    #     name="fc" + str(classes),
    #     kernel_initializer=glorot_uniform(seed=0),
    # )(X)

    # Create model
    model = tf.keras.Model(inputs=X_input, outputs=X, name="ResNet50")

    return model


def identity_block(X, f, filters, stage, block):
    """
    Implementation of the identity block

    Arguments:
    X -- input tensor of shape (m, n_H_prev, n_W_prev, n_C_prev)
    f -- integer, specifying the shape of the middle CONV's window for the main path
    filters -- python list of integers, defining the number of filters in the CONV layers of the main path
    stage -- integer, used to name the layers, depending on their position in the network
    block -- string/character, used to name the layers, depending on their position in the network

    Returns:
    X -- output of the identity block, tensor of shape (n_H, n_W, n_C)
    """

    # defining name basis
    conv_name_base = "res" + str(stage) + block + "_branch"
    bn_name_base = "bn" + str(stage) + block + "_branch"

    # Retrieve Filters
    F1, F2, F3 = filters

    # Save the input value. You'll need this later to add back to the main path.
    X_shortcut = X

    # First component of main path
    X = tf.keras.layers.Conv2D(
        filters=F1,
        kernel_size=(1, 1),
        strides=(1, 1),
        padding="valid",
        name=conv_name_base + "2a",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X)
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2a")(X)
    X = tf.keras.layers.Activation("relu")(X)

    # Second component of main path
    X = tf.keras.layers.Conv2D(
        filters=F2,
        kernel_size=(f, f),
        strides=(1, 1),
        padding="same",
        name=conv_name_base + "2b",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X)
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2b")(X)
    X = tf.keras.layers.Activation("relu")(X)

    # Third component of main path
    X = tf.keras.layers.Conv2D(
        filters=F3,
        kernel_size=(1, 1),
        strides=(1, 1),
        padding="valid",
        name=conv_name_base + "2c",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X)
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2c")(X)

    # Final step: Add shortcut value to main path, and pass it through a RELU activation
    X = tf.keras.layers.Add()([X, X_shortcut])
    X = tf.keras.layers.Activation("relu")(X)

    return X


def convolutional_block(X, f, filters, stage, block, s=2):
    """
    Implementation of the convolutional block

    Arguments:
    X -- input tensor of shape (m, n_H_prev, n_W_prev, n_C_prev)
    f -- integer, specifying the shape of the middle CONV's window for the main path
    filters -- python list of integers, defining the number of filters in the CONV layers of the main path
    stage -- integer, used to name the layers, depending on their position in the network
    block -- string/character, used to name the layers, depending on their position in the network
    s -- Integer, specifying the stride to be used

    Returns:
    X -- output of the convolutional block, tensor of shape (n_H, n_W, n_C)
    """

    # defining name basis
    conv_name_base = "res" + str(stage) + block + "_branch"
    bn_name_base = "bn" + str(stage) + block + "_branch"

    # Retrieve Filters
    F1, F2, F3 = filters

    # Save the input value
    X_shortcut = X

    ##### MAIN PATH #####
    # First component of main path
    X = tf.keras.layers.Conv2D(
        F1,
        (1, 1),
        strides=(s, s),
        name=conv_name_base + "2a",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X)
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2a")(X)
    X = tf.keras.layers.Activation("relu")(X)

    # Second component of main path
    X = tf.keras.layers.Conv2D(
        F2,
        (f, f),
        strides=(1, 1),
        padding="same",
        name=conv_name_base + "2b",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X)
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2b")(X)
    X = tf.keras.layers.Activation("relu")(X)

    # Third component of main path
    X = tf.keras.layers.Conv2D(
        F3,
        (1, 1),
        strides=(1, 1),
        padding="valid",
        name=conv_name_base + "2c",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X)
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2c")(X)

    ##### SHORTCUT PATH ####
    X_shortcut = tf.keras.layers.Conv2D(
        F3,
        (1, 1),
        strides=(s, s),
        padding="valid",
        name=conv_name_base + "1",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
    )(X_shortcut)
    X_shortcut = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "1")(
        X_shortcut
    )

    # Final step: Add shortcut value to main path, and pass it through a RELU activation
    X = tf.keras.layers.Add()([X, X_shortcut])
    X = tf.keras.layers.Activation("relu")(X)

    return X



import cv2
import numpy as np


class Kalman:
    def __init__(self):
        self.reset_kalman()

    def reset_kalman(self):
        self.kalman = cv2.KalmanFilter(4, 2)
        self.kalman.measurementMatrix = np.eye(2, 4, dtype=np.float32)

        self.kalman.transitionMatrix = np.array(
            [[1, 0, 1, 0], [0, 1, 0, 1], [0, 0, 1, 0], [0, 0, 0, 1]], np.float32
        )

        self.kalman.processNoiseCov = np.eye(4, 4, dtype=np.float32) * 0.03

    def predict(self):
        return self.kalman.predict()

    def correct(self, rect):
        pts = np.array(
            [np.float32(rect.centroid[0]), np.float32(rect.centroid[1])], np.float32
        )
        self.kalman.correct(pts)



"""
# Download latest model relase from  https://github.com/TheCacophonyProject/AI-Model
# with prefix server-
# place this in /etc/cacophony/classifier.yaml
"""

import sys
import requests
from pathlib import Path
import subprocess
import yaml
import json


def get_headers():
    return {
        "Accept": "application/vnd.github+json",
        "X-GitHub-Api-Version": "2022-11-28",
    }


def run_command(cmd, timeout=None):
    proc = subprocess.run(
        cmd,
        shell=True,
        encoding="ascii",
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        timeout=timeout,
    )
    return proc.stdout


def get_releases():
    r = requests.get(
        "https://api.github.com/repos/TheCacophonyProject/AI-Model/releases",
        headers=get_headers(),
    )
    r.raise_for_status()

    return r.json()


def download_file(url, name, extract_to):
    print(f"Downloading {url}")
    response = requests.get(url)
    open(name, "wb").write(response.content)
    out_dir = Path("latest-model")
    out_dir.mkdir(exist_ok=True)
    print(f"Extracting {name} to {extract_to}")
    extract_to.mkdir()
    run_command(f"tar -xzf {name} -C {extract_to} --strip-components=1")
    Path(name).unlink()


def main():
    all_releases = get_releases()
    server_releases = [
        release
        for release in all_releases
        if release["tag_name"].startswith("server") and not release["prerelease"]
    ]
    rf_releases = [
        release
        for release in all_releases
        if release["tag_name"].startswith("rf-fp") and not release["prerelease"]
    ]

    rf_releases = sorted(rf_releases, key=lambda r: r["tag_name"], reverse=True)
    server_releases = sorted(server_releases, key=lambda r: r["tag_name"], reverse=True)
    if len(rf_releases) == 0 and len(server_releases) == 0:
        print("No releases found for server")
        sys.exit(0)

    config_file = "/etc/cacophony/classifier.yaml"
    print("Loading ", config_file)
    with open(config_file) as stream:
        config = yaml.safe_load(stream)

    if config is None:
        config = {"classify": {"models": []}}

    add_to_config(server_releases[0], config, 1, "Keras")
    add_to_config(rf_releases[0], config, 2, "RandomForest")

    with open(config_file, "w") as stream:
        raw = yaml.dump(config, stream)


def add_to_config(release, config, model_id, model_type):
    print("Using release ", release["tag_name"])
    if len(release["assets"]) == 0:
        print("Release has no files", release)
        return False
    asset = release["assets"][0]

    # check existing models
    existing_models = Path("/etc/cacophony/models")
    model_versions = []
    if existing_models.exists():
        for dir in existing_models.iterdir():
            if dir.is_dir():
                model_versions.append(dir.name)
    else:
        existing_models.mkdir()
    print("Have existing", model_versions)
    if release["tag_name"] in model_versions:
        print("Already have latest model")
        return False

    # Download and extract new model
    print("Downloading model", asset["name"])
    model_dir = existing_models / release["tag_name"]
    download_file(asset["browser_download_url"], asset["name"], model_dir)

    # Get model metadata to extract model name
    meta_data = model_dir.glob("*.json")
    print("looking for meta in ", "latest-model")
    meta_data = next(meta_data)
    print("Loading metadata ", meta_data)
    with open(meta_data) as stream:
        model_meta = json.load(stream)
    model_name = model_meta["name"]
    print("Model name is ", model_name)

    # if config.get("classify") is None:
    #     config["classify"] = {}
    # if len(config.get("classify", {}).get("models", [])) == 0:
    #     config_model = {
    #         "name": model_name,
    #         "model_file": str(model_dir / "saved_model.pb"),
    #         "id": model_id,
    #     }
    #     config["classify"]["models"] = [config_model]
    # else:
    config_model = config["classify"]["models"][model_id - 1]
    config_model["name"] = model_name
    if model_type == "RandomForest":
        config_model["model_file"] = str(meta_data.with_suffix(".pkl"))
    else:
        config_model["model_file"] = str(model_dir / "saved_model.pb")
    config_model["id"] = config_model["id"] + 1
    config_model["type"] = model_type

    return True


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print("Error running ", e)
        raise e



# import argparse

from cptv import CPTVReader
from config.config import Config
from ml_tools.logs import init_logging
from piclassifier.cptvmotiondetector import CPTVMotionDetector
from config.thermalconfig import ThermalConfig
from piclassifier.headerinfo import HeaderInfo

#
#
# def parse_args():
#     parser = argparse.ArgumentParser()
#     parser.add_argument(
#         "cptv", default="clips/possum.cptv", help="a CPTV file to detect motion"
#     )
#     parser.add_argument(
#         "-c",
#         "--config-file",
#         default="tests/config.toml",
#         help="Path to config file to use",
#     )
#     parser.add_argument(
#         "--thermal-config-file",
#         default="test-config.yaml",
#         help="Path to pi-config file (config.toml) to use",
#     )
#
#     args = parser.parse_args()
#     return args


def test_motion(
    cptv_file="tests/clips/possum.cptv",
    thermal_config_file="tests/config.toml",
    config_file="tests/test-config.yaml",
):
    from pathlib import Path

    config = Config.load_from_file(config_file)
    thermal_config = ThermalConfig.load_from_file(thermal_config_file)
    print("detecting on  " + cptv_file)

    with open(cptv_file, "rb") as f:
        reader = CPTVReader(f)

        headers = HeaderInfo(
            res_x=reader.x_resolution,
            res_y=reader.y_resolution,
            fps=9,
            brand="",
            model="",
            frame_size=reader.x_resolution * reader.y_resolution * 2,
            pixel_bits=16,
            serial="",
            firmware="",
        )

        motion_detector = CPTVMotionDetector(
            thermal_config, config.tracking["thermal"].motion.dynamic_thresh, headers
        )
        for i, frame in enumerate(reader):
            motion_detector.process_frame(frame)



"""
Tools for visualising ML models.

"""

import numpy as np
import matplotlib.pyplot as plt
import itertools


def show_saliency_map(model, X_in, y_in):
    """
    Plots saliency map for single segment via pyplot.
    :param model: the model to use to generate the saliency maps
    :param X_in: segment of shape [frames, height, width, channels]
    :param y_in: label index
    """

    X = np.asarray(X_in, dtype=np.float32)
    y = np.asarray(y_in, dtype=np.int32)

    saliency = compute_saliency_map(X[np.newaxis, :, :, :, :], y[np.newaxis], model)[0]
    saliency = saliency / (np.max(saliency) + 1e-8)

    cols = X.shape[0]

    rows = 8

    for frame in range(cols):
        # plot original image

        # plot all 4 channels plus original
        for channel in range(4):
            plt.subplot(rows, cols, (cols * channel) + frame + 1)
            frame_data = np.float32(X[frame, channel, :, :])
            if channel in [2, 3]:
                # for motion vectors it's better to use magnitude when drawing them
                frame_data = np.abs(frame_data)

            plt.imshow(frame_data, aspect="auto", vmin=-1, vmax=10)
            plt.axis("off")

            plt.subplot(rows, cols, (cols * (channel + 4)) + frame + 1)
            plt.imshow(
                saliency[frame, channel, :, :],
                vmin=0.0,
                vmax=1.0,
                cmap=plt.cm.hot,
                aspect="auto",
            )
            plt.axis("off")

    plt.subplots_adjust(wspace=0, hspace=0)
    plt.gcf().set_size_inches(cols * 3, rows * 3)
    plt.show()


def show_segment(X_in):
    """
    Displays all channels of a segment
    :param X_in: segment of shape [frames, height, width, channels]
    """

    X = np.asarray(X_in, dtype=np.float32)

    cols = X.shape[0]

    rows = 4

    for frame in range(cols):
        # plot original image

        # plot all 4 channels plus original
        for channel in range(4):
            plt.subplot(rows, cols, (cols * channel) + frame + 1)
            frame_data = np.float32(X[frame, channel, :, :])
            if channel in [2, 3]:
                # for motion vectors it's better to use magnitude when drawing them
                frame_data = np.abs(frame_data) * 5

            plt.imshow(frame_data, aspect="auto", vmin=-10, vmax=10)
            plt.axis("off")

    plt.subplots_adjust(wspace=0, hspace=0)
    plt.gcf().set_size_inches(cols * 3, rows * 3)
    plt.show()


def plot_confidence_by_class(predictions, true_class, labels):
    """
    Plots confidence for each class label for both correct and incorrect predictions.
    :param predictions: prediction distribution for each example. Of shape [n, classes]
    :param true_class: true class index for each example.  Of shape [n]
    :param labels: names of labels for each class.  List of strings of length [n]
    :return: a figure
    """
    class_conf_correct = []
    class_conf_incorrect = []
    for i in range(len(labels)):
        class_conf_correct.append([])
        class_conf_incorrect.append([])

    pred_class = [np.argmax(prediction) for prediction in predictions]
    pred_conf = [predictions[i][x] for i, x in enumerate(pred_class)]

    for i, conf in enumerate(pred_conf):
        if true_class[i] == pred_class[i]:
            class_conf_correct[true_class[i]].append(conf)
        else:
            class_conf_incorrect[pred_class[i]].append(conf)

    fig = plt.figure(1, figsize=(8, 6))
    ax = plt.gca()

    plt.grid(True)
    ax.set_axisbelow(True)

    y_pos = np.arange(len(labels))
    bar_width = 0.35

    values = [np.mean(class_conf_correct[i]) for i in range(len(labels))]
    r1 = plt.bar(y_pos, values, bar_width, alpha=0.9, label="Correct")

    values = [np.mean(class_conf_incorrect[i]) for i in range(len(labels))]
    r2 = plt.bar(y_pos + bar_width, values, bar_width, alpha=0.9, label="Incorrect")

    plt.xticks(y_pos, labels)
    plt.ylabel("Confidence")
    plt.title("Confidence by Class on Correct Segments.")
    ax.set_ylim([0.0, 1.0])
    plt.legend()

    plt.tight_layout()
    return fig


def fig_to_numpy(figure):
    """Converts a matplotlib figure to a numpy array."""
    figure.canvas.draw()
    data = figure.canvas.tostring_rgb()
    ncols, nrows = figure.canvas.get_width_height()
    return np.fromstring(data, dtype=np.uint8).reshape(nrows, ncols, 3)


def plot_confusion_matrix(confusion_matrix, classes):
    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111)
    ax.matshow(confusion_matrix, cmap=plt.cm.Blues)

    plt.title("Classification Confusion Matrix")

    thresh = confusion_matrix.max() / 2.0
    for i, j in itertools.product(
        range(confusion_matrix.shape[0]), range(confusion_matrix.shape[1])
    ):
        plt.text(
            j,
            i,
            "{:.2f}".format(confusion_matrix[i, j]),
            horizontalalignment="center",
            color="white" if confusion_matrix[i, j] > thresh else "black",
        )

    ax.set_xticklabels([""] + classes, rotation=45)
    ax.set_yticklabels([""] + classes)
    ax.xaxis.set_tick_params(labeltop=False, labelbottom=True)
    return fig


def compute_saliency_map(X_in, y_in, model):
    """
    Compute a class saliency map for segment using the model for image X and label y.
    :param X_in: segment of shape [batch, frames, channels, height, width]
    :param y_in: label index
    :param model: the model to use classify the segment
    :return: the saliency map of shape [frames, height, width]
    """

    global tf
    import tensorflow as tf

    correct_scores = tf.gather_nd(
        model.prediction,
        tf.stack((tf.range(X_in.shape[0], dtype="int64"), model.y), axis=1),
    )

    feed_dict = model.get_feed_dict(X_in, y_in)

    print(correct_scores.shape, model.X.shape)

    grads = tf.abs(tf.gradients(correct_scores, model.X)[0])
    saliency = model.session.run([grads], feed_dict=feed_dict)[0]

    return saliency



"""This script handles the training of models. It works in two different modes.

# Single model training

run the script with
$ train.py [model_name]

to train a single model. Hyper-parameters are read from classifier.yaml.

# Search mode

Many models will be trained across a set of predified hyper
parameters.  This can take days, but gives useful information.  I have
two dictionaries, a SHORT_SEARCH_PARAMS which has just a few
interesting parameters to try, and FULL_SEARCH_PARAMS which is more
comprehensive.

The results of the search are stored in "search_results.txt" in the
"train" subdirectory of `base_data_folder`. Jobs that have already
been processed will not be redone (however cancelling partway through
a job will cause it restart from the start of the job).

# Checking the results

All the training results are stored in tensorboard.  To assess the
training run tensorboard from the log directory.

"""

import argparse
import os

import matplotlib

matplotlib.use("Agg")  # enable canvas drawing

from config.config import Config
from train.train import train_model


def load_config():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config-file", help="Path to config file to use")
    parser.add_argument("-g", "--grid", action="count", help="Grid Search hparams")
    parser.add_argument("-w", "--weights", help="Fine tune using these weights")
    parser.add_argument("-i", "--ignore", help="Ignore clips in this file")
    parser.add_argument("-e", "--epochs", type=int, help="Epochs to train")
    parser.add_argument("-f", "--fine_tune", help="Model to fine tune")

    parser.add_argument(
        "name",
        default="unnammed",
        help="Name of training job",
    )
    args = parser.parse_args()
    return Config.load_from_file(args.config_file), args


def main():
    conf, args = load_config()

    os.makedirs(conf.train.train_dir, exist_ok=True)
    train_model(
        args.name,
        conf,
        conf.train.hyper_params,
        do_grid_search=args.grid,
        weights=args.weights,
        ignore=args.ignore,
        epochs=args.epochs,
        fine_tune=args.fine_tune,
    )


if __name__ == "__main__":
    main()



from datetime import datetime, timedelta
import json
import logging
import os
import time
import psutil
import numpy as np
import logging
from track.clip import Clip
from track.track import ThumbInfo

from .motiondetector import SlidingWindow
from .processor import Processor

from ml_tools.logs import init_logging
from ml_tools.rectangle import Rectangle
from . import beacon

from piclassifier.trapcontroller import trigger_trap
from piclassifier.attiny import set_recording_state

SNAPSHOT_SIGNAL = "snap"
STOP_SIGNAL = "stop"
SKIP_SIGNAL = "skip"
track_extractor = None
clip = None
fp_model = None
classifier = None


def run_classifier(
    frame_queue, config, thermal_config, headers, classify=True, detect_after=None
):
    init_logging()
    pi_classifier = None
    try:
        pi_classifier = PiClassifier(
            config, thermal_config, headers, classify, detect_after
        )
        while True:
            frame = frame_queue.get()
            if isinstance(frame, str):
                if frame == STOP_SIGNAL:
                    logging.info("PiClassifier received stop signal")
                    pi_classifier.disconnected()
                    return
                if frame == "skip":
                    pi_classifier.skip_frame()
                elif frame == SNAPSHOT_SIGNAL:
                    pi_classifier.take_snapshot()
            else:
                pi_classifier.process_frame(frame[0], frame[1])
    except:
        logging.error("Error running classifier restarting ..", exc_info=True)
        if pi_classifier is not None:
            pi_classifier.disconnected()
        return


predictions = None
use_low_power_mode = False


class PiClassifier(Processor):
    """Classifies frames from leptond"""

    NUM_CONCURRENT_TRACKS = 1
    DEBUG_EVERY = 20
    MAX_CONSEC = 1
    # after every MAX_CONSEC frames skip this many frames
    # this gives the cpu a break

    # try classify a non fp track every X frames
    SKIP_FRAMES = 25
    # only do another full classification on the same track after this many frames
    PREDICT_EVERY = 40

    # run fp model predictions every X frames
    FP_MODEL_SKIP_FRAMES = 10
    # only do another fp classification on the same track after this many frames
    FP_PREDICT_EVERY = 30

    def __init__(
        self,
        config,
        thermal_config,
        headers,
        classify,
        detect_after=None,
        preview_type=None,
    ):
        self.constant_recorder = None
        self._output_dir = thermal_config.recorder.output_dir
        self.headers = headers
        self.classifier = None
        self.fp_model = None
        self.frame_num = 0
        self.clip = None
        self.prev_clip = None
        self.enable_per_track_information = False
        self.rolling_track_classify = {}
        self.next_classify_frame = 0
        self.next_fp_classification_frame = 0
        self.classified_consec = 0
        self.classify = classify
        self.config = config
        self.predictions = {}
        self.process_time = 0
        self.tracking_time = 0
        self.identify_time = 0
        self.fp_identify_time = 0
        self.total_time = 0
        self.rec_time = 0
        self.fp_time = 0
        self.monitored_tracks = {}
        self.recording = False
        self.tracking_events = thermal_config.motion.tracking_events
        self.bluetooth_beacons = thermal_config.motion.bluetooth_beacons
        self.preview_frames = thermal_config.recorder.preview_secs * headers.fps
        self.do_tracking = thermal_config.motion.do_tracking
        self.fps_timer = SlidingWindow((headers.fps * 3), np.float32)
        self.preview_type = preview_type
        self.max_keep_frames = None if preview_type else 0
        self.track_extractor = None
        self.use_low_power_mode = thermal_config.recorder.use_low_power_mode
        global use_lower_power_mode
        use_low_power_mode = self.use_low_power_mode
        if not use_low_power_mode:
            # clear state
            set_recording_state(False)
        if thermal_config.recorder.disable_recordings:
            from .dummyrecorder import DummyRecorder

            self.recorder = DummyRecorder(
                thermal_config, headers, on_recording_stopping=on_recording_stopping
            )

        if headers.model == "IR":
            from track.irtrackextractor import IRTrackExtractor
            from .irmotiondetector import IRMotionDetector

            logging.info("Running on IR")
            PiClassifier.SKIP_FRAMES = 3

            if self.do_tracking:

                self.track_extractor = IRTrackExtractor(
                    self.config.tracking,
                    cache_to_disk=self.config.classify.cache_to_disk,
                    keep_frames=False,
                    verbose=config.verbose,
                    calc_stats=False,
                    scale=0.25,
                    on_trapped=on_track_trapped,
                    update_background=False,
                    trap_size=thermal_config.device_setup.trap_size,
                    from_pi=True,
                )
            self.tracking_config = self.config.tracking.get(IRTrackExtractor.TYPE)

            self.type = IRTrackExtractor.TYPE
            from .irrecorder import IRRecorder

            if not thermal_config.recorder.disable_recordings:
                self.recorder = IRRecorder(
                    thermal_config, headers, on_recording_stopping=on_recording_stopping
                )
            self.snapshot_recorder = IRRecorder(
                thermal_config,
                headers,
                on_recording_stopping=on_recording_stopping,
                name="IR Snapshot",
            )
            self.motion_detector = IRMotionDetector(
                thermal_config,
                headers,
            )
            if thermal_config.recorder.constant_recorder:
                self.constant_recorder = IRRecorder(
                    thermal_config,
                    headers,
                    on_recording_stopping=on_recording_stopping,
                    name="IR Constant",
                    constant_recorder=True,
                )
        else:
            logging.info("Running on Thermal")
            from .cptvmotiondetector import CPTVMotionDetector

            if self.do_tracking:
                from track.cliptrackextractor import ClipTrackExtractor

                self.track_extractor = ClipTrackExtractor(
                    self.config.tracking,
                    self.config.use_opt_flow,
                    self.config.classify.cache_to_disk,
                    keep_frames=False,
                    calc_stats=False,
                    update_background=False,
                    from_pi=True,
                )
            self.tracking_config = self.config.tracking.get("thermal")

            self.type = "thermal"
            if not thermal_config.recorder.disable_recordings:
                from .cptvrecorder import CPTVRecorder

                self.recorder = CPTVRecorder(
                    thermal_config, headers, on_recording_stopping=on_recording_stopping
                )
            self.snapshot_recorder = CPTVRecorder(
                thermal_config,
                headers,
                on_recording_stopping=on_recording_stopping,
                constant_recorder=False,
                name="CPTV Snapshot",
                file_suffix="-snap",
            )
            self.motion_detector = CPTVMotionDetector(
                thermal_config,
                self.tracking_config.motion.dynamic_thresh,
                headers,
                detect_after=detect_after,
            )
            if thermal_config.recorder.constant_recorder:
                self.constant_recorder = CPTVRecorder(
                    thermal_config,
                    headers,
                    on_recording_stopping=on_recording_stopping,
                    name="CPTV Constant",
                    constant_recorder=True,
                )
        edge = self.tracking_config.edge_pixels
        self.crop_rectangle = Rectangle(
            edge, edge, headers.res_x - 2 * edge, headers.res_y - 2 * edge
        )
        global track_extractor
        track_extractor = self.track_extractor
        self.motion = thermal_config.motion
        self.min_frames = thermal_config.recorder.min_secs * headers.fps
        self.max_frames = thermal_config.recorder.max_secs * headers.fps
        if thermal_config.recorder.disable_recordings:
            from .dummyrecorder import DummyRecorder

            self.recorder = DummyRecorder(
                thermal_config, headers, on_recording_stopping=on_recording_stopping
            )
        if thermal_config.throttler.activate:
            from .throttledrecorder import ThrottledRecorder

            self.recorder = ThrottledRecorder(
                self.recorder,
                thermal_config,
                headers,
                on_recording_stopping=on_recording_stopping,
            )
        self.meta_dir = os.path.join(thermal_config.recorder.output_dir)
        if not os.path.exists(self.meta_dir):
            os.makedirs(self.meta_dir)

        self.max_keep_frames = 25
        self.max_pred_frames = None
        if self.classify and self.do_tracking:
            from ml_tools.interpreter import get_interpreter
            from classify.trackprediction import Predictions

            model = None
            fp_config = None
            for model_config in config.classify.models:
                if model_config.type != "RandomForest":
                    model = model_config
                else:
                    fp_config = model_config

            if model is not None:
                self.classifier = get_interpreter(model)
                global classifier
                classifier = self.classifier
                self.frames_per_classify = (
                    self.classifier.params.square_width
                    * self.classifier.params.square_width
                )
                if self.frames_per_classify > 1:
                    self.predict_from_last = self.frames_per_classify * 2

                self.max_keep_frames = (
                    self.frames_per_classify * 2 if not preview_type else None
                )
                self.predictions[model.id] = Predictions(self.classifier.labels, model)
                self.num_labels = len(self.classifier.labels)
                logging.info("Labels are %s ", self.classifier.labels)
                global predictions
                predictions = self.predictions
                try:
                    self.fp_index = self.classifier.labels.index("false-positive")
                except ValueError:
                    self.fp_index = None
                self.startup_classifier()
            if fp_config is not None:
                self.fp_model = get_interpreter(fp_config)
                global fp_model
                fp_model = self.fp_model
                self.predictions[self.fp_model.id] = Predictions(
                    self.fp_model.labels, fp_config
                )

        super().__init__()

    def new_clip(self, preview_frames):
        self.clip = Clip(
            self.tracking_config,
            "stream",
            model=self.headers.model,
            type=self.type,
            calc_stats=False,
            fps=self.headers.fps,
        )
        global clip
        clip = self.clip
        self.clip.video_start_time = datetime.now() - timedelta(
            seconds=len(preview_frames) / self.headers.fps
        )

        self.clip.num_preview_frames = len(preview_frames)
        self.next_classify_frame = 0
        self.next_fp_classification_frame = 0
        self.clip.set_res(self.res_x, self.res_y)
        self.clip.set_frame_buffer(
            self.tracking_config.high_quality_optical_flow,
            self.config.classify.cache_to_disk,
            self.config.use_opt_flow,
            keep_frames=(
                self.max_keep_frames > 0 if self.max_keep_frames is not None else True
            ),
            max_frames=self.max_keep_frames,
        )

        self.clip.update_background(self.motion_detector.background.copy())
        self.clip._background_calculated()
        if self.do_tracking == False:
            return
        # no need to retrack all of preview
        background_frames = None
        track_frames = -1
        retrack_back = True
        if self.type == "IR":
            track_frames = 5
            retrack_back = True
            # background is calculated in motion, so already 5 frames ahead
        self.track_extractor.start_tracking(
            self.clip,
            preview_frames,
            track_frames=track_frames,
            background_alg=self.motion_detector._background,
            retrack_back=retrack_back,
            # background_frame=clip.background,
            # background_frames=background_frames,
        )

    def startup_classifier(self):
        # classifies an empty frame to force loading of the model into memory
        num_inputs, in_shape = self.classifier.shape()
        if num_inputs > 1:
            zero_input = []
            for shape in in_shape:
                zero_input.append(np.zeros((1, *shape[1:]), np.float32))
        else:
            zero_input = np.zeros((1, *in_shape[1:]), np.float32)
        self.classifier.predict(zero_input)

    def get_active_tracks(self):
        """
        Gets current clips active_tracks and returns the top NUM_CONCURRENT_TRACKS order by priority
        """
        active_tracks = self.clip.active_tracks
        active_tracks = [track for track in active_tracks if len(track) >= 8]
        return active_tracks
        filtered = []
        for track in active_tracks:
            pred = None
            if self.predictions is not None:
                pred = self.predictions.prediction_for(track.get_id())
            if pred is not None:
                if (
                    pred.last_frame_classified is not None
                    and self.clip.current_frame - pred.last_frame_classified
                    < PiClassifier.PREDICT_EVERY
                ):
                    logging.debug(
                        "Skipping %s as predicted %s and now at %s",
                        track,
                        pred.last_frame_classified,
                        self.clip.current_frame,
                    )
                    continue

            filtered.append(track)
        active_tracks = filtered
        if (
            len(active_tracks) <= PiClassifier.NUM_CONCURRENT_TRACKS
            or not self.classify
        ):
            return active_tracks
        active_predictions = []
        for track in active_tracks:
            prediction = self.predictions.get_or_create_prediction(track, keep_all=True)
            active_predictions.append(prediction)

        top_priority = sorted(
            active_predictions,
            key=lambda i: i.get_priority(self.clip.current_frame),
            reverse=True,
        )

        top_priority = [
            track.track_id
            for track in top_priority[: PiClassifier.NUM_CONCURRENT_TRACKS]
        ]
        classify_tracks = [
            track for track in active_tracks if track.get_id() in top_priority
        ]
        return classify_tracks

    def identify_last_frame(self):
        """
        Runs through track identifying segments, and then returns it's prediction of what kind of animal this is.
        One prediction will be made for every active_track of the last frame.
        :return: TrackPrediction object
        """

        prediction_smooth = 0.1

        smooth_prediction = None
        smooth_novelty = None

        prediction = 0.0
        novelty = 0.0
        if (
            self.next_fp_classification_frame >= self.clip.current_frame
            and self.next_classify_frame >= self.clip.current_frame
        ):
            return

        self.next_fp_classification_frame += PiClassifier.FP_MODEL_SKIP_FRAMES
        active_tracks = self.get_active_tracks()
        new_prediction = False
        if len(active_tracks) == 0:
            return False

        if self.fp_model is not None:
            fp_time = time.time()
            for track in active_tracks:
                start = time.time()
                if self.classifier is not None:
                    full_model = self.predictions[self.classifier.id].prediction_for(
                        track.get_id()
                    )
                    if full_model is not None and full_model.num_frames_classified > 0:
                        logging.debug(
                            "Skipping fp model for %s as has full model prediction",
                            track,
                        )
                        continue
                track_prediction = self.predictions[
                    self.fp_model.id
                ].get_or_create_prediction(track, keep_all=True)
                if (
                    track_prediction.last_frame_classified is not None
                    and self.clip.current_frame - track_prediction.last_frame_classified
                    < PiClassifier.FP_PREDICT_EVERY
                ):
                    logging.debug(
                        "Skipping %s #%s last %s",
                        track,
                        self.clip.current_frame,
                        track_prediction.last_frame_classified,
                    )
                    continue
                result = self.fp_model.predict_track(
                    clip,
                    track,
                    predict_from_last=45,
                    max_frames=PiClassifier.FP_PREDICT_EVERY // 5,
                    num_predictions=1,
                    last_frame_predicted=track_prediction.last_frame_classified,
                )
                if result is None:
                    track_prediction.last_frame_classified = self.clip.current_frame
                    continue
                frames, prediction, mass = result
                if prediction is None:
                    track_prediction.last_frame_classified = self.clip.current_frame
                    continue
                for p, m, frame in zip(prediction, mass, frames):
                    track_prediction.classified_frame(frame, p, m)

                logging.debug(
                    "Track %s is predicted as %s took %s track frames %s",
                    track,
                    track_prediction.get_prediction(),
                    time.time() - start,
                    len(track),
                )
                new_prediction = True
            self.fp_identify_time += time.time() - fp_time
        if (
            self.classifier is not None
            and self.next_classify_frame <= self.clip.current_frame
        ):
            id_start = time.time()

            self.next_classify_frame += PiClassifier.SKIP_FRAMES
            animal_tracks = self.get_active_animal_tracks_for_predicting()
            # filter based of fp model
            for i, track in enumerate(animal_tracks):
                logging.debug("Running full classifier on %s", track)
                track_prediction = self.predictions[
                    self.classifier.id
                ].get_or_create_prediction(track, keep_all=True)
                start = time.time()
                frames, prediction, mass = self.classifier.predict_track(
                    clip,
                    track,
                    predict_from_last=self.predict_from_last,
                    scale=self.track_extractor.scale,
                    frames_per_classify=self.frames_per_classify,
                    max_frames=self.max_pred_frames,
                    num_predictions=1,
                    calculate_filtered=True,
                    last_frame_predicted=track_prediction.last_frame_classified,
                )

                if prediction is None:
                    track_prediction.last_frame_classified = self.clip.current_frame
                    continue
                for p, m in zip(prediction, mass):
                    track_prediction.classified_frames(frames, p, m)

                logging.info(
                    "Track %s is predicted as %s took %s track frames %s",
                    track,
                    track_prediction.get_prediction(),
                    time.time() - start,
                    len(track),
                )
                new_prediction = True

            self.identify_time += time.time() - id_start

        for i, track in enumerate(active_tracks):
            if self.tracking_events:
                track_prediction, model_id = self.get_best_prediction(track.get_id())

                if track_prediction.predicted_tag() != "false-positive":
                    track_prediction.tracking = True
                    self.monitored_tracks[track.get_id()] = track
                elif track_prediction.tracking:
                    # tracking ended as is false-positive
                    track_prediction.tracking = False
                    track_prediction.normalize_score()
                    self.service.tracking(
                        self.clip._id,
                        track,
                        track_prediction.class_best_score,
                        track.bounds_history[-1],
                        False,
                        track_prediction.last_frame_classified,
                        self.predictions[model_id].labels,
                        model_id,
                    )
                    if track.get_id() in self.monitored_tracks:
                        del self.monitored_tracks[track.get_id()]

        if self.bluetooth_beacons:
            if new_prediction:
                active_predictions = []
                for track in self.clip.active_tracks:
                    track_prediction, model_id = self.get_best_prediction(
                        track.get_id()
                    )
                    if track_prediction:
                        active_predictions.append(track_prediction)
                beacon.classification(active_predictions)
        return new_prediction

    def get_active_animal_tracks_for_predicting(self):
        active_tracks = self.get_active_tracks()
        filtered = []
        least_fp_track = None
        for track in active_tracks:
            if self.fp_model is not None:
                pred, model_id = self.get_best_prediction(track.get_id())
                logging.debug(
                    "track %s -%s - %s",
                    track.get_id(),
                    pred.predicted_tag(),
                    pred.normalized_best_score(),
                )
                if pred is not None:
                    if pred.predicted_tag() == "false-positive":
                        confidence = pred.normalized_best_score()

                        if confidence >= 0.7:
                            if least_fp_track is None or least_fp_track[0] > confidence:
                                least_fp_track = (confidence, track)
                            logging.debug(
                                "Skipping track %s as is FP confidence %s",
                                track,
                                confidence,
                            )
                            continue

            pred = None
            if self.predictions is not None:
                pred = self.predictions[self.classifier.id].prediction_for(
                    track.get_id()
                )
            if pred is not None:
                if len(pred.predictions) < 2:
                    classify_every = PiClassifier.PREDICT_EVERY // 2
                else:
                    classify_every = PiClassifier.PREDICT_EVERY
                if (
                    pred.last_frame_classified is not None
                    and self.clip.current_frame - pred.last_frame_classified
                    < classify_every
                ):
                    logging.debug(
                        "Skipping %s as predicted %s and now at %s",
                        track,
                        pred.last_frame_classified,
                        self.clip.current_frame,
                    )
                    continue

            filtered.append(track)

        active_tracks = filtered
        if len(active_tracks) == 0:
            if least_fp_track is None:
                return []
            logging.debug("Using least fp track %s", least_fp_track[1])
            return [least_fp_track[1]]

        # choose most likely animals first
        active_tracks = sorted(
            active_tracks,
            key=lambda track: self.animal_ranking(track),
            reverse=True,
        )
        if len(active_tracks) > PiClassifier.NUM_CONCURRENT_TRACKS:
            active_tracks = active_tracks[: PiClassifier.NUM_CONCURRENT_TRACKS]
        return active_tracks

    def animal_ranking(self, track):
        track_pred, model_id = self.get_best_prediction(track.get_id())

        if track_pred is None or track_pred.class_best_score is None:
            return 0
        fp_confidence = track_pred.class_best_score[track_pred.fp_index] / np.sum(
            track_pred.class_best_score
        )
        return 1 - fp_confidence

    def get_thumbnail(self, clip_id=None, track_id=None):
        import cv2
        from ml_tools.imageprocessing import resize_and_pad

        start = time.time()
        if clip_id is not None:
            if self.prev_clip is not None and self.prev_clip._id == clip_id:
                if track_id is not None:
                    best_track = next(
                        [
                            track
                            for track in self.prev_clip.tracks
                            if track.get_id() == track_id
                        ],
                        default=None,
                    )
                    if best_track is None:
                        logging.info("Couldn't find track %s", track_id)
                        return None
                else:
                    best_track = None
                    for track in tracks:
                        if track.thumb_info is not None:
                            if (
                                best_track is None
                                or track.thumb_info.score()
                                > best_track.thumb_info.score()
                            ):
                                best_track = track
                return (
                    best_track.thumb_info.thumb,
                    best_track.thumb_info.track_id,
                    best_track.thumb_info.region,
                )

            elif self.clip is not None and self.clip._id != clip_id:
                logging.info("Cant find requested clip id %s", clip_id)
                return None
        elif self.clip is None:
            logging.info("Have no clip")
            return None
        if clip.frame_buffer.frames is None or len(clip.frame_buffer.frames) == 0:
            logging.info("Have no frames")
            return None
        with clip.frame_buffer.frame_lock:
            tracks = clip.active_tracks
            oldest_frame_num = clip.frame_buffer.frames[0].frame_number
            best_contour = None
            if track_id is not None:
                tracks = [track for track in tracks if track.get_id() == track_id]
                if len(tracks) == 0:
                    logging.info("Couldn't find track %s", track.get_id())
                    return None

            for track in tracks:
                confidence = None
                tag = None
                if predictions is not None:
                    pred, model_id = self.get_best_prediction(track.get_id())
                    if pred is not None and pred.max_score is not None:
                        confidence = round(100 * pred.max_score)
                        tag = pred.predicted_tag()
                regions = track.bounds_history
                if track.thumb_info is None:
                    track.thumb_info = ThumbInfo(track.get_id())
                track.thumb_info.predicted_confidence = confidence
                track.thumb_info.predicted_tag = tag
                # if (
                #     track.thumb_info.region is not None
                #     and track.thumb_info.region.frame_number >= regions[-1].frame_number
                # ):
                #     # logging.info("Already up to date for track %s", track.get_id())
                #     continue
                i = len(regions) - 1
                first_loop = True
                while i >= 0:
                    region = regions[i]
                    if (
                        track.thumb_info.last_frame_check is not None
                        and track.thumb_info.last_frame_check >= region.frame_number
                    ):
                        break
                    if region.frame_number >= oldest_frame_num:
                        frame = clip.frame_buffer.frames[
                            region.frame_number - oldest_frame_num
                        ]

                        if first_loop:
                            track.thumb_info.last_frame_check = region.frame_number
                        first_loop = False
                        assert frame.frame_number == region.frame_number
                        contour_image = (
                            frame.filtered if frame.mask is None else frame.mask
                        )
                        contours, _ = cv2.findContours(
                            np.uint8(region.subimage(contour_image)),
                            cv2.RETR_EXTERNAL,
                            # cv2.CHAIN_APPROX_SIMPLE,
                            cv2.CHAIN_APPROX_TC89_L1,
                        )

                        contour_points = 0
                        if len(contours) > 0:
                            contour_points = len(contours[0])
                        if track.thumb_info.points < contour_points:
                            track.thumb_info.points = contour_points
                            track.thumb_info.region = region
                            track.thumb_info.thumb = None
                    else:
                        break
                    i -= 1
                if (
                    best_contour is None
                    or track.thumb_info.score() > best_contour.score()
                    or (
                        track.thumb_info.predicted_tag != "false-positive"
                        and best_contour.predicted_tag == "false-positive"
                    )
                ):
                    best_contour = track.thumb_info

            if track_id is not None:
                track = tracks[0]
                best_contour = track.thumb_info
            if best_contour is None:
                return None
            for track in tracks:
                if track.thumb_info.region is None:
                    continue
                thumb_frame = track.thumb_info.region.frame_number
                if (
                    track.thumb_info.thumb is None
                    or thumb_frame > track.thumb_info.thumb_frame
                ):
                    frame = clip.frame_buffer.frames[thumb_frame - oldest_frame_num]
                    thumb_thermal = track.thumb_info.region.subimage(frame.thermal)
                    if thumb_thermal.shape[0] > 32 or thumb_thermal.shape[1] > 32:
                        thumb_thermal = resize_and_pad(
                            thumb_thermal, (32, 32), None, None
                        )
                    track.thumb_info.thumb = np.uint16(thumb_thermal)
                    track.thumb_info.thumb_frame = thumb_frame
            # logging.info("Took %s to get thumbnail", time.time() - start)
            return best_contour.thumb, best_contour.track_id, best_contour.region

    def get_recent_frame(self, last_frame=None):
        # save us having to lock if we dont have a different frame
        if last_frame is not None and self.motion_detector.num_frames == last_frame:
            return None, None, last_frame
        last_frame = self.motion_detector.get_recent_frame()
        if self.clip:
            if last_frame is None:
                return None
            track_meta = []
            tracks = clip.active_tracks
            for track in tracks:
                pred = None
                if self.predictions:
                    pred = {
                        self.predictions[self.classifier.id].model.id: self.predictions[
                            self.classifier.id
                        ]
                    }
                meta = track.get_metadata(pred)
                last_pos = meta["positions"][-1].copy()
                # if self.track_extractor.scale is not None:
                # last_pos.rescale(1 / self.track_extractor.scale)
                meta["positions"] = [last_pos]
                track_meta.append(meta)

            return last_frame, track_meta, self.motion_detector.num_frames
        else:
            return (
                last_frame,
                {},
                self.motion_detector.num_frames,
            )

    def disconnected(self):
        self.motion_detector.disconnected()
        if self.recorder.recording and self.tracking_events:
            self.recording = False
            self.service.recording(False)
        self.recorder.force_stop()
        self.snapshot_recorder.force_stop()
        if self.constant_recorder is not None:
            self.constant_recorder.force_stop()

        self.end_clip()
        self.service.quit()

    def skip_frame(self):
        if self.clip:
            self.clip.current_frame += 1

    def take_snapshot(self):
        started = self.snapshot_recorder.start_recording(
            None, [], self.motion_detector.temp_thresh, time.time()
        )
        if not started:
            logging.info("Already taking snapshot recording")
            return False
        logging.info("Taking new snapshot recorder")
        self.snapshot_recorder.write_until = 2 * self.headers.fps
        return True

    def process_frame(self, lepton_frame, received_at):
        import time

        start = time.time()
        self.motion_detector.process_frame(lepton_frame)
        self.process_time += time.time() - start
        if self.snapshot_recorder.recording:
            self.snapshot_recorder.process_frame(False, lepton_frame, received_at)
        if self.constant_recorder is not None and self.motion_detector.can_record():
            if self.constant_recorder.recording:
                self.constant_recorder.process_frame(True, lepton_frame, received_at)
            else:
                logging.info("Starting new constant recorder")
                self.recording = self.constant_recorder.start_recording(
                    self.motion_detector.background,
                    [],
                    self.motion_detector.temp_thresh,
                    time.time(),
                )
                if self.recording and not self.use_low_power_mode:
                    set_recording_state(True)
        if (
            not self.recorder.recording
            and self.motion_detector.movement_detected
            and not lepton_frame.ffc_imminent
            and not lepton_frame.ffc_status in [1, 2]
        ):
            s_r = time.time()
            preview_frames = self.motion_detector.preview_frames()
            bak = self.motion_detector.background
            self.recording = self.recorder.start_recording(
                self.motion_detector.background,
                preview_frames,
                self.motion_detector.temp_thresh,
                received_at,
            )
            self.rec_time += time.time() - s_r
            if self.recording:
                if self.tracking_events:
                    self.service.recording(True)
                if not self.use_low_power_mode:
                    set_recording_state(True)

                if self.bluetooth_beacons:
                    beacon.recording()
                t_start = time.time()
                self.new_clip(preview_frames)
                self.tracking_time += time.time() - t_start

        if self.recorder.recording:
            t_start = time.time()
            if self.do_tracking:
                self.track_extractor.process_frame(self.clip, lepton_frame)
                self.get_thumbnail()
            self.tracking_time += time.time() - t_start
            s_r = time.time()

            self.recorder.process_frame(
                self.motion_detector.movement_detected, lepton_frame, received_at
            )
            self.rec_time += time.time() - s_r
            if self.classify:
                if self.motion_detector.calibrating:
                    # dont think we will get ffcs if we are recording
                    self.classified_consec = 0
                else:
                    identified = self.identify_last_frame()
                    if not identified:
                        self.classified_consec = 0
            elif len(self.monitored_tracks) == 0 and self.tracking_events:
                active_tracks = self.get_active_tracks()

                active_tracks = [
                    track
                    for track in active_tracks
                    if len(track) > 10 and track.last_bound.mass > 16
                ]
                logging.debug(
                    "got active tracks bigger than 16 and longer than 10 frames %s",
                    active_tracks,
                )

                active_tracks = sorted(
                    active_tracks,
                    key=lambda track: track.last_mass,
                    reverse=True,
                )

                if len(active_tracks) > 0:
                    logging.debug(
                        "tracking by biggest mass %s",
                        active_tracks[0],
                    )
                    self.monitored_tracks[active_tracks[0].get_id()] = active_tracks[0]

            if len(self.monitored_tracks) > 0:
                monitored_tracks = list(self.monitored_tracks.values())
                for monitored_track in monitored_tracks:
                    tracking = monitored_track in self.clip.active_tracks
                    score = 0
                    prediction = ""
                    all_scores = None
                    model_id = 0
                    track_prediction = None
                    last_prediction = 0
                    if self.classify:
                        track_prediction, model_id = self.get_best_prediction(
                            monitored_track.get_id()
                        )
                        all_scores = track_prediction.get_normalized_score()
                        last_prediction = track_prediction.last_frame_classified
                    self.service.tracking(
                        self.clip._id,
                        monitored_track,
                        all_scores,
                        monitored_track.bounds_history[-1],
                        tracking,
                        last_prediction,
                        [] if model_id is None else self.predictions[model_id].labels,
                        model_id,
                    )

                    if not tracking:
                        del self.monitored_tracks[monitored_track.get_id()]
                        if self.classify:
                            track_prediction.tracking = False
        elif self.clip is not None:
            self.end_clip()

        if not self.recorder.recording and self.recording and self.tracking_events:
            self.recording = False
            self.service.recording(False)

        self.frame_num += 1
        self.total_time += time.time() - start
        if (
            self.motion_detector.can_record()
            and self.frame_num % PiClassifier.DEBUG_EVERY == 0
        ):
            if self.clip is not None:
                average = np.mean(self.fps_timer.get_frames())
                mem = process_mem()
                logging.debug(
                    "tracking %s %% process %s %%  identify %s %% FP Id %s %%  rec %s %% fps %s/sec process  system cpu %s process memory %s%% system memory %s behind by %s seconds",
                    round(100 * self.tracking_time / self.total_time, 3),
                    round(100 * self.process_time / self.total_time, 3),
                    round(100 * self.identify_time / self.total_time, 3),
                    round(100 * self.fp_identify_time / self.total_time, 3),
                    round(100 * self.rec_time / self.total_time, 3),
                    round(1 / average),
                    psutil.cpu_percent(),
                    mem,
                    psutil.virtual_memory()[2],
                    time.time() - received_at,
                )
            self.tracking_time = 0
            self.process_time = 0
            self.identify_time = 0
            self.fp_identify_time = 0
            self.total_time = 0
            self.rec_time = 0
        self.fps_timer.add(time.time() - start)

    def create_mp4(self):
        from ml_tools.previewer import Previewer

        previewer = Previewer(self.config, self.preview_type)
        previewer.export_clip_preview(
            os.path.join(self.output_dir, self.clip.get_id() + ".mp4"),
            self.clip,
            self.predictions if self.classify else None,
        )

    def get_best_prediction(self, track_id):
        if self.classifier is not None:
            main_classifier = self.predictions[self.classifier.id].prediction_for(
                track_id
            )
            if (
                main_classifier is not None
                and main_classifier.num_frames_classified > 0
            ):
                return main_classifier, self.classifier.id
        if self.fp_model is not None:
            return (
                self.predictions[self.fp_model.id].prediction_for(track_id),
                self.fp_model.id,
            )
        return None, None

    def end_clip(self):
        if self.clip:
            global clip
            if self.preview_type:
                self.create_mp4()
            logging.debug(
                "Ending clip with %s tracks post filtering", len(self.clip.tracks)
            )
            if self.classify:
                for pred in self.predictions.values():
                    logging.info("Pred for model %s", pred.model)
                    for t_id, prediction in pred.prediction_per_track.items():
                        if prediction.max_score:
                            logging.info(
                                "Clip {} {} {}".format(
                                    self.clip.get_id(),
                                    t_id,
                                    prediction.description(),
                                )
                            )
                    pred.clear_predictions()
            # set so we can get thumbnail info
            self.prev_clip = clip
            self.prev_clip.frame_buffer = None
            self.clip = None
            self.monitored_tracks = {}
            clip = None

    @property
    def res_x(self):
        return self.headers.res_x

    @property
    def res_y(self):
        return self.headers.res_y

    @property
    def output_dir(self):
        return self._output_dir


def on_track_trapped(track):
    track.trap_reported = True
    # GP could make a prediction here

    global predictions

    tag = None
    if predictions is not None:
        pred = predictions.prediction_for(track.get_id())
        if pred is not None:
            tag = pred.predicted_tag()
            track.trap_tag = tag
    logging.warn("Trapped track %s with tag %s", track, tag)
    trigger_trap(tag)


def on_recording_stopping(filename):
    from ml_tools.tools import CustomJSONEncoder

    global use_low_power_mode
    if not use_low_power_mode:
        set_recording_state(False)

    if "-snap" in filename.stem:
        return
    global clip, track_extractor, predictions
    if clip and track_extractor:
        track_extractor.apply_track_filtering(clip)
        if predictions is not None:
            valid_preds = {}

            # remove track prediction
            for track in clip.tracks:
                for model_pred in predictions.values():
                    pred = model_pred.prediction_for(track.get_id)
                    if pred is not None:
                        pred.normalize_score()
            #             valid_preds[model_pred.model.id] = pred
            # for model_id in predictions.keys():
            #     if model_id in valid_preds:
            #         predictions[model_id].prediction_per_track = valid_preds[model_id].prediction_per_track

        # filter criteria has been scaled so resize after
        # if track_extractor.scale is not None:
        #     for track in clip.tracks:
        #         for r in track.bounds_history:
        #             # bring back to orignal size
        #             r.rescale(1 / track_extractor.scale)

        meta_name = os.path.splitext(filename)[0] + ".txt"
        logging.debug("saving meta to %s", meta_name)
        predictions_per_model = None

        if predictions is not None:
            predictions_per_model = predictions
        meta_data = clip.get_metadata(predictions_per_model)
        meta_data["algorithm"] = {}
        meta_data["algorithm"]["tracker_version"] = f"PI-{track_extractor.VERSION}"
        meta_data["metadata_source"] = "PI"
        if predictions is not None:
            models = []
            model_name = ""
            joiner = ""
            for model_preds in predictions.values():
                models.append(model_preds.model.as_dict())
                model_name = f"{model_name}{joiner}{model_preds.model.name}"
                joiner = " and "
            meta_data["algorithm"]["model_name"] = model_name
            meta_data["models"] = models

        with open(meta_name, "w") as f:
            json.dump(meta_data, f, indent=4, cls=CustomJSONEncoder)


def process_mem():
    # return the memory usage in percentage like top
    process = psutil.Process(os.getppid())
    return process.memory_percent()



import matplotlib.pyplot as plt

import tensorflow as tf
from functools import partial
import numpy as np
import time
import math
import logging
from config.config import Config
import json

import sys
from ml_tools.logs import init_logging

AUTOTUNE = tf.data.AUTOTUNE
# IMAGE_SIZE = [256, 256]
# BATCH_SIZE = 64


def get_excluded():
    return ["human", "dog", "nothing", "other", "sheep", "unknown", "rodent"]


def get_remapped():
    return {"insect": "false-positive", "cat": "possum"}
    # return ["insect", "cat"]


def load_dataset(filenames, remap_lookup, labels, args):
    deterministic = args.get("deterministic", False)

    ignore_order = tf.data.Options()
    ignore_order.experimental_deterministic = (
        deterministic  # disable order, increase speed
    )
    dataset = tf.data.TFRecordDataset(
        filenames, compression_type="GZIP", num_parallel_reads=4
    )
    dataset = dataset.with_options(
        ignore_order
    )  # uses data as soon as it streams in, rather than in its original order

    image_size = args["image_size"]
    labeled = args.get("labeled", True)
    augment = args.get("augment", False)
    preprocess_fn = args.get("preprocess_fn")
    one_hot = args.get("one_hot", True)
    dataset = dataset.apply(tf.data.experimental.ignore_errors())
    dataset = dataset.map(
        partial(
            read_irrecord,
            remap_lookup=remap_lookup,
            num_labels=len(labels),
            image_size=image_size,
            labeled=labeled,
            augment=augment,
            preprocess_fn=preprocess_fn,
            one_hot=one_hot,
            include_track=args.get("include_track", False),
        ),
        num_parallel_calls=AUTOTUNE,
        deterministic=deterministic,
    )
    if args.get("include_track", False):
        filter_excluded = lambda x, y: not tf.math.equal(tf.math.count_nonzero(y[0]), 0)
    else:
        filter_excluded = lambda x, y: not tf.math.equal(tf.math.count_nonzero(y), 0)

    dataset = dataset.filter(filter_excluded)

    return dataset


data_augmentation = tf.keras.Sequential(
    [
        tf.keras.layers.RandomFlip("horizontal"),
        tf.keras.layers.RandomRotation(0.1, fill_mode="nearest", fill_value=0),
        tf.keras.layers.RandomZoom(0.1),
        tf.keras.layers.RandomBrightness(0.2),
        tf.keras.layers.RandomContrast(0.1),
    ]
)


def read_irrecord(
    example,
    image_size,
    remap_lookup,
    num_labels,
    labeled,
    augment,
    preprocess_fn=None,
    one_hot=True,
    include_track=False,
):
    tfrecord_format = {
        "image/augmented": tf.io.FixedLenFeature((), tf.int64, 0),
        "image/thermalencoded": tf.io.FixedLenFeature((), tf.string),
        "image/filteredencoded": tf.io.FixedLenFeature((), tf.string),
        "image/height": tf.io.FixedLenFeature((), tf.int64, -1),
        "image/width": tf.io.FixedLenFeature((), tf.int64, -1),
        "image/class/label": tf.io.FixedLenFeature((), tf.int64, -1),
    }
    if include_track:
        tfrecord_format["image/track_id"] = tf.io.FixedLenFeature((), tf.int64, -1)
        tfrecord_format["image/avg_mass"] = tf.io.FixedLenFeature((), tf.int64, -1)

    example = tf.io.parse_single_example(example, tfrecord_format)
    image = decode_image(
        example["image/thermalencoded"],
        example["image/filteredencoded"],
        image_size,
        augment,
    )
    augmented = tf.cast(example["image/augmented"], tf.bool)

    if not augmented and augment:
        image = data_augmentation(image)
    if preprocess_fn is not None:
        image = preprocess_fn(image)

    if labeled:
        label = tf.cast(example["image/class/label"], tf.int32)
        label = remap_lookup.lookup(label)
        if one_hot:
            label = tf.one_hot(label, num_labels)
        if include_track:
            track_id = tf.cast(example["image/track_id"], tf.int32)
            avg_mass = tf.cast(example["image/avg_mass"], tf.int32)
            label = (label, track_id, avg_mass)
        return image, label
    return image


def decode_image(image, filtered, image_size, augment):
    image = tf.image.decode_png(image, channels=1)
    image = tf.concat((image, image), axis=2)

    image = tf.cast(image, tf.float32)
    image = tf.image.resize_with_pad(image, image_size[0], image_size[1])
    return image


def main():
    init_logging()

    from .tfdataset import get_dataset, get_distribution

    config = Config.load_from_file("classifier-ir.yaml")

    file = f"{config.base_folder}/training-data/training-meta.json"
    with open(file, "r") as f:
        meta = json.load(f)
    labels = meta.get("labels", [])
    datasets = []
    # weights = [0.5] * len(labels)
    resampled_ds, remapped, labels, _ = get_dataset(
        load_dataset,
        f"{config.base_folder}/training-data/test",
        labels,
        batch_size=1,
        image_size=(160, 160),
        augment=False,
        excluded_labels=get_excluded(),
        remapped_labels=get_remapped(),
        deterministic=True,
        shuffle=False,
    )
    meta["remapped"] = remapped
    with open(file, "w") as f:
        json.dump(meta, f)
    from collections import Counter
    import cv2

    i = 0
    for e in range(2):
        # for batch_x, batch_y in resampled_ds:
        #     for x, y in zip(batch_x, batch_y):
        #         lbl = np.argmax(y)
        #         lbl = labels[lbl]
        #         print("X is", x.shape, lbl)
        #         cv2.imwrite(f"./images/{lbl}-{i}.png", x.numpy())
        #         # 1 / 0
        #         i += 1
        # # for x_2 in x:
        # #     print("max is", np.amax(x_2), x_2.shape)
        # #     assert np.amax(x_2) == 255
        # # show_batch(x, y, labels)
        # return
        print("epoch", e)
        dist = get_distribution(resampled_ds, len(labels), extra_meta=False)
        #
        # true_categories = tf.concat([y for x, y in resampled_ds], axis=0)
        # true_categories = np.int64(tf.argmax(true_categories, axis=1))
        # c = Counter(list(true_categories))
        # print("epoch is size", len(true_categories))
        for l, d in zip(labels, dist):
            print("after have", l, d)


def show_batch(image_batch, label_batch, labels):
    plt.figure(figsize=(10, 10))
    print("images in batch", len(image_batch))
    num_images = min(len(image_batch), 25)
    for n in range(num_images):
        print("imageas max is", np.amax(image_batch[n]), np.amin(image_batch[n]))
        # print("image bach", image_batch[n])
        ax = plt.subplot(5, 5, n + 1)
        plt.imshow(np.uint8(image_batch[n]))
        plt.title(labels[np.argmax(label_batch[n])])
        plt.axis("off")
    plt.show()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Take in a single optional integral argument
import sys

DBUS_NAME = "org.cacophony.thermalrecorder"
DBUS_PATH = "/org/cacophony/thermalrecorder"
# Create a reference to the RandomData object on the  session bus
from gi.repository import GLib

import dbus
import dbus.mainloop.glib
import time
import threading
from datetime import datetime
import cv2
import numpy as np

model_labels = []

active_tracks = {}


def catchall_tracking_signals_handler(
    clip_id,
    track_id,
    prediction,
    what,
    confidence,
    region,
    frame,
    mass,
    blank,
    tracking,
    last_prediction_frame,
    model_id,
):
    print(
        f"Received a tracking signal for clip {clip_id} track {track_id} and it says {what} {confidence}% at {region} tracking? {tracking} prediction {prediction} frame {frame} mass {mass} blank? {blank} last predicted frame {last_prediction_frame}"
    )
    index = 0
    print("Model is ", model_id)
    labels = model_labels[model_id]
    for x in prediction:
        print("For  ", labels[index], " have confidence ", int(x), "%")
        index += 1

    if tracking:
        if track_id not in active_tracks:
            active_tracks[track_id] = True
            print("Tracking", track_id)
    else:
        # active_tracks[track_id] = False
        print("Stopped tracking", track_id)

    print("Active tracks are ", active_tracks.keys())
    bus = dbus.SystemBus()
    dbus_object = bus.get_object(DBUS_NAME, DBUS_PATH)
    thumb, track_id, region = dbus_object.GetThumbnail(clip_id, track_id)
    thumb = np.uint16(thumb)
    thumb = normalize(thumb)
    cv2.imshow("t", thumb)
    cv2.waitKey(100)


def normalize(thumb):
    a_max = np.amax(thumb)
    a_min = np.amin(thumb)
    return np.uint8(255 * (np.float32(thumb) - a_min) / (a_max - a_min))


def catchall_rec_signals_handler(dt, is_recording):
    if is_recording:
        print("Recording started at ", datetime.fromtimestamp(dt))
    else:
        print("Recording ended at ", datetime.fromtimestamp(dt))


# helper class to run dbus in background
class TrackingService:
    def __init__(self, callback, rec_callback):
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self.callback = callback
        self.rec_callback = rec_callback
        self.loop = GLib.MainLoop()
        self.t = threading.Thread(
            target=self.run_server,
        )
        self.t.start()

    def quit(self):
        self.loop.quit()

    def run_server(self):
        dbus_object = None
        try:
            bus = dbus.SystemBus()
            dbus_object = bus.get_object(DBUS_NAME, DBUS_PATH)
        except dbus.exceptions.DBusException as e:
            print("Failed to initialize D-Bus object: '%s'" % str(e))
            sys.exit(2)
        global model_labels
        model_labels = dbus_object.ClassificationLabels()

        bus.add_signal_receiver(
            self.callback,
            dbus_interface=DBUS_NAME,
            signal_name="Tracking",
        )
        bus.add_signal_receiver(
            self.rec_callback,
            dbus_interface=DBUS_NAME,
            signal_name="Recording",
        )
        self.loop.run()


if __name__ == "__main__":
    tracking = TrackingService(
        catchall_tracking_signals_handler, catchall_rec_signals_handler
    )

    # just to keep program alive
    while tracking.t.is_alive():
        time.sleep(1)



import sys
import time
import logging
from inotify_simple import INotify, flags


def monitor_file(on_modify, filename="/etc/cacophony/config.toml"):
    inotify = INotify()
    watch_flags = flags.CREATE | flags.MODIFY | flags.DELETE_SELF
    wd = inotify.add_watch("/etc/cacophony/config.toml", watch_flags)
    for event in inotify.read():
        event_reasons = [str(flag) for flag in flags.from_mask(event.mask)]
        on_modify(",".join(event_reasons))



import numpy as np

from ml_tools.rectangle import Rectangle


class TestRectangle:
    def test_can_create_rectangle_from_width_and_height(self):
        rectangle = Rectangle(2, 3, 5, 6)
        assert_rectangle_values(rectangle)

    def test_crop(self):
        rectangle = Rectangle(0, 0, 100, 100)
        rectangle.crop(Rectangle(2, 3, 5, 6))

        assert_rectangle_values(rectangle)

    def test_subimage(self):
        image = np.arange(100).reshape((10, 10))
        rectangle = Rectangle(2, 3, 2, 3)

        subimage = rectangle.subimage(image)
        assert np.array_equal(subimage, [[32, 33], [42, 43], [52, 53]])


def assert_rectangle_values(rect):
    assert rect.left == 2
    assert rect.top == 3
    assert rect.width == 5
    assert rect.height == 6



import logging
from piclassifier.recorder import Recorder


class DummyRecorder(Recorder):
    def __init__(self, thermal_config, headers, on_recording_stopping=None):
        self.recording = False
        self.frames = 0
        self.min_frames = thermal_config.recorder.min_secs * headers.fps
        self.max_frames = thermal_config.recorder.max_secs * headers.fps

    def process_frame(self, movement_detected, cptv_frame, received_at):
        if self.recording:
            self.write_frame(cptv_frame)
            if movement_detected:
                self.write_until = self.frames + self.min_frames
            elif self.has_minimum():
                self.stop_recording(received_at)
                return

            if self.frames == self.max_frames:
                self.stop_recording(received_at)
        return

    def has_minimum(self):
        return self.frames > self.write_until

    def write_frame(self, cptv_frame):
        self.frames += 1

    def start_recording(
        self, background_frame, preview_frames, temp_thresh, frame_time
    ):
        if self.recording:
            return False
        if not self.can_record(frame_time):
            return False
        self.recording = True
        self.frames = 0
        for f in preview_frames:
            self.write_frame(f)
        self.write_until = self.frames + self.min_frames
        logging.info("Dummy recording started temp_thresh: %s", temp_thresh)

        return True

    def stop_recording(self, frame_time):
        self.recording = False
        self.write_until = 0
        logging.info("Dummy recording ended  %s frames", self.frames)

    def force_stop(self):
        self.recording = False
        self.write_until = 0



import numpy as np
from datetime import timedelta
import logging

from .motiondetector import SlidingWindow, MotionDetector, WeightedBackground
from ml_tools.rectangle import Rectangle


class CPTVMotionDetector(MotionDetector):
    FFC_PERIOD = timedelta(seconds=9.9)
    BACKGROUND_WEIGHT_ADD = 0.1

    def __init__(self, thermal_config, dynamic_thresh, headers, detect_after=None):
        super().__init__(thermal_config, headers)
        self.headers = headers
        if headers.model and headers.model.lower() == "lepton3.5":
            CPTVMotionDetector.BACKGROUND_WEIGHT_ADD = 1
        self.config = thermal_config.motion
        self.location_config = thermal_config.location
        self.num_preview_frames = thermal_config.recorder.preview_secs * headers.fps
        self.compare_gap = self.config.frame_compare_gap + 1
        edge = self.config.edge_pixels
        self.min_frames = thermal_config.recorder.min_secs * headers.fps
        self.max_frames = thermal_config.recorder.max_secs * headers.fps

        if not self.config.one_diff_only:
            self.diff_window = SlidingWindow(
                self.compare_gap,
                np.int32,
            )
        self.running_mean = None
        self.running_mean_frames = 0
        self.thermal_window = SlidingWindow(self.num_preview_frames + 1, "O")
        self.processed = 0
        self.thermal_thresh = 0
        self.crop_rectangle = Rectangle(
            edge, edge, headers.res_x - 2 * edge, headers.res_y - 2 * edge
        )
        self._background = WeightedBackground(
            edge,
            self.crop_rectangle,
            self.res_x,
            self.res_y,
            CPTVMotionDetector.BACKGROUND_WEIGHT_ADD,
            self.config.temp_thresh,
        )
        self.movement_detected = False
        self.dynamic_thresh = dynamic_thresh

        self.triggered = 0

        self.ffc_affected = False
        if detect_after is None:
            self.detect_after = self.thermal_window.size * 2
        else:
            self.detect_after = detect_after

    @property
    def calibrating(self):
        return self.ffc_affected

    def preview_frames(self):
        return self.thermal_window.get_frames()[:-1]

    @property
    def temp_thresh(self):
        return self._background.average

    def detect(self, clipped_frame, received_at=None):
        oldest = self.crop_rectangle.subimage(self.thermal_window.oldest_nonffc.pix)
        oldest = np.clip(oldest, a_min=self.temp_thresh, a_max=None)
        clipped_frame = np.clip(clipped_frame, a_min=self.temp_thresh, a_max=None)
        delta_frame = clipped_frame - oldest
        if not self.config.warmer_only:
            delta_frame = abs(delta_frame)
        if self.config.one_diff_only:
            diff = len(delta_frame[delta_frame > self.config.delta_thresh])
        else:
            if self.processed > 2:
                delta_frame2 = self.diff_window.oldest_nonffc
                delta_frame[delta_frame >= self.config.delta_thresh] = (
                    self.config.delta_thresh
                )
                delta_combined = delta_frame2 + delta_frame
                diff = len(
                    delta_combined[delta_combined == self.config.delta_thresh * 2]
                )
            else:
                delta_frame[delta_frame >= self.config.delta_thresh] = (
                    self.config.delta_thresh
                )
                diff = 0

            self.diff_window.add(delta_frame, self.ffc_affected)
        if diff > self.config.count_thresh:
            if not self.movement_detected:
                logging.debug(
                    "{} MotionDetector motion detected {} thresh {} count {} ".format(
                        timedelta(seconds=self.num_frames / 9),
                        self.processed,
                        self.temp_thresh,
                        diff,
                    )
                )
            return True

        if self.movement_detected:
            logging.debug(
                "{} MotionDetector motion stopped thresh {} count {}".format(
                    timedelta(seconds=self.num_frames / 9),
                    self.temp_thresh,
                    diff,
                )
            )
        return False

    @property
    def background(self):
        return self._background.background

    def get_recent_frame(self):
        return self.thermal_window.current

    def disconnected(self):
        self.thermal_window.reset()
        if not self.config.one_diff_only:
            self.diff_window.reset()
        self.processed = 0

    def process_frame(self, cptv_frame, force_process=False):
        prev_ffc = self.ffc_affected
        self.ffc_affected = is_affected_by_ffc(cptv_frame)
        if self.can_record() or force_process:
            oldest_thermal = self.thermal_window.oldest_nonffc
            self.thermal_window.add(cptv_frame, self.ffc_affected)
            cropped_frame = np.int32(self.crop_rectangle.subimage(cptv_frame.pix))
            if not self.ffc_affected:
                if self.running_mean is None:
                    frames = self.thermal_window.get_frames()[:45]
                    last_45 = [f.pix for f in frames if not is_affected_by_ffc(f)]
                    if len(last_45) > 0:
                        self.running_mean_frames = len(last_45)
                        self.running_mean = np.mean(last_45, axis=0)
                else:
                    oldest_thermal = oldest_thermal.pix
                    if self.running_mean_frames == 45:
                        self.running_mean -= oldest_thermal / self.running_mean_frames
                        self.running_mean += cptv_frame.pix / self.running_mean_frames
                    else:
                        self.running_mean = (
                            self.running_mean * self.running_mean_frames
                            + cptv_frame.pix
                        )
                        self.running_mean_frames += 1
                        self.running_mean /= self.running_mean_frames
                if self.running_mean is not None:
                    self._background.process_frame(self.running_mean)
            if self.ffc_affected or prev_ffc:
                logging.debug("{} MotionDetector FFC".format(self.num_frames))
                self.movement_detected = False
                self.triggered = 0
                if prev_ffc:
                    self.thermal_window.non_ffc_index = self.thermal_window.last_index
            elif self.processed > self.detect_after:
                movement = self.detect(cropped_frame)
                if movement:
                    self.triggered += 1
                else:
                    self.triggered = 0
                if self.triggered >= self.config.trigger_frames:
                    self.movement_detected = True
                else:
                    self.movement_detected = False
            self.processed += 1
        else:
            self.thermal_window.update_current_frame(cptv_frame, self.ffc_affected)
            self.movement_detected = False

        self.num_frames += 1
        return self.movement_detected

    def skip_frame(self):
        return


def is_affected_by_ffc(cptv_frame):
    if hasattr(cptv_frame, "ffc_status") and cptv_frame.ffc_status in [1, 2]:
        return True

    if cptv_frame.time_on is None or cptv_frame.last_ffc_time is None:
        return False
    if isinstance(cptv_frame.time_on, int):
        return (
            cptv_frame.time_on - cptv_frame.last_ffc_time
        ) < CPTVMotionDetector.FFC_PERIOD.seconds
    return (
        cptv_frame.time_on - cptv_frame.last_ffc_time
    ) < CPTVMotionDetector.FFC_PERIOD



import math
import numpy as np
from track.region import Region
import cv2
import matplotlib as mpl
import matplotlib.pyplot as plt
from collections import namedtuple
from ml_tools import tools


def best_trackless_thumb(clip):
    """Choose a frame for clips without any track"""
    best_region = None
    THUMBNAIL_SIZE = 64

    # if we have regions take best mass of un tracked regions
    for regions in clip.region_history:
        for region in regions:
            if best_region is None or region.mass > best_region.mass:
                best_region = region
    if best_region is not None:
        return best_region

    # take region with greatest filtered mean values, and
    # if zero take thermal mean values
    best_frame_i = np.argmax(clip.stats.frame_stats_mean)
    best_frame = clip.frame_buffer.get_frame(best_frame_i).thermal
    frame_height, frame_width = best_frame.shape
    best_filtered = best_frame - clip.background
    best_region = None
    for y in range(frame_height - THUMBNAIL_SIZE):
        for x in range(frame_width - THUMBNAIL_SIZE):
            thermal_sum = np.mean(
                best_frame[y : y + THUMBNAIL_SIZE, x : x + THUMBNAIL_SIZE]
            )
            filtered_sum = np.mean(
                best_filtered[y : y + THUMBNAIL_SIZE, x : x + THUMBNAIL_SIZE]
            )
            if best_region is None:
                best_region = ((x, y), filtered_sum, thermal_sum)
            elif best_region[1] > 0:
                if best_region[1] < filtered_sum:
                    best_region = ((x, y), thermal_sum, filtered_sum)
            elif best_region[2] < thermal_sum:
                best_region = ((x, y), thermal_sum, filtered_sum)
    centroid = (
        best_region[0][0] + THUMBNAIL_SIZE // 2,
        best_region[0][1] + THUMBNAIL_SIZE // 2,
    )
    return Region(
        best_region[0][0],
        best_region[0][1],
        THUMBNAIL_SIZE,
        THUMBNAIL_SIZE,
        frame_number=best_frame_i,
        centroid=centroid,
    )


Stat = namedtuple("Stat", "region contours median_diff")


def get_track_thumb_stats(clip, track):
    max_mass = 0
    max_median_diff = 0
    min_median_diff = 0
    max_contour = 0
    stats = []
    for region in track.bounds_history:
        if region.blank or region.mass == 0:
            continue
        frame = clip.frame_buffer.get_frame(region.frame_number)
        if frame is None:
            continue
        contour_image = frame.filtered if frame.mask is None else frame.mask
        contours, _ = cv2.findContours(
            np.uint8(region.subimage(contour_image)),
            cv2.RETR_EXTERNAL,
            # cv2.CHAIN_APPROX_SIMPLE,
            cv2.CHAIN_APPROX_TC89_L1,
        )
        if len(contours) == 0:
            # shouldnt happen
            # contour_points.append(0)
            continue
        else:
            points = len(contours[0])
            if points > max_contour:
                max_contour = points

        # get mask of all pixels that are considered an animal
        filtered_sub = region.subimage(contour_image)
        sub_mask = filtered_sub > 0

        # get the thermal values for this mask
        thermal_sub = region.subimage(frame.thermal)
        masked_thermal = thermal_sub[sub_mask]

        # get the difference in media between this and the frame median
        t_median = np.median(frame.thermal)
        masked_median = np.median(masked_thermal)
        median_diff = masked_median - t_median

        if region.mass > max_mass:
            max_mass = region.mass
        if median_diff > max_median_diff:
            max_median_diff = median_diff
        if median_diff < min_median_diff:
            min_median_diff = median_diff
        stats.append(Stat(region, points, median_diff))
    return stats, max_mass, max_median_diff, min_median_diff, max_contour


def get_thumbnail_info(clip, track):
    (
        stats,
        max_mass,
        max_median_diff,
        min_median_diff,
        max_contour,
    ) = get_track_thumb_stats(clip, track)
    if len(stats) == 0:
        if len(track.bounds_history) == 0:
            return None, 0
        return Stat(track.bounds_history[0], 0, 0), 0
    scored_frames = sorted(
        stats,
        key=lambda s: score(s, max_mass, max_median_diff, min_median_diff, max_contour),
        reverse=True,
    )

    best_score = score(
        scored_frames[0], max_mass, max_median_diff, min_median_diff, max_contour
    )
    return scored_frames[0], best_score


def score(stat, max_mass, max_median_diff, min_median_diff, max_contour):
    region = stat.region
    # mass out of 40
    mass_percent = region.mass / max_mass
    mass_percent = mass_percent * 40

    # contours out of 50
    pts = stat.contours / max_contour
    pts = pts * 50

    centroid_mid = tools.eucl_distance_sq(region.centroid, region.mid) ** 0.5
    centroid_mid *= 2
    # this will be probably between 1-10 could be worth * 1.5
    # median diff out of 50
    if max_median_diff == 0:
        # not sure to zero or what here
        diff = 0
        if min_median_diff != 0:
            diff = stat.median_diff + abs(min_median_diff)
            diff = diff / abs(min_median_diff)
            diff = diff * 40
    else:
        diff = stat.median_diff / max_median_diff
        diff = diff * 40

    score = mass_percent + pts + diff - centroid_mid

    is_along_border = (
        region.x <= 1 or region.y <= 1 or region.bottom >= 119 or region.right >= 159
    )
    # prefer frames not on border
    if is_along_border:
        score = score - 1000
    return score


# just for testing
def display_track(h_data, id):
    rows = len(h_data)
    columns = 10
    fig = plt.figure(figsize=(50, 50))

    fig.suptitle(
        f"{id}  Suggested thumbnails",
        fontsize=16,
    )
    offset = 0
    for heruistic in h_data:
        for i, h_info in enumerate(heruistic):
            plt.subplot(rows, columns, i + 1 + offset)
            frame = h_info[0]
            stat = h_info[1]
            region = stat.region
            is_along_border = (
                region.x <= 1
                or region.y <= 1
                or region.bottom >= 119
                or region.right >= 159
            )
            centroid_mid = tools.eucl_distance_sq(region.centroid, region.mid) ** 0.5
            title = f"#{region.frame_number} - centroid {round(centroid_mid)} - {is_along_border}\n"
            title += f" cts {stat.contours}"
            title += f" mass {region.mass}"
            title += f" diff {stat.median_diff}"

            remove_axes(title)

            plt.imshow(region.subimage(frame.thermal))
        offset += columns

    plt.subplots_adjust(0, 0, 0.99, 0.99, 0, 0)
    plt.show()
    plt.close(fig)
    plt.close()


def remove_axes(title):
    ax = plt.gca()
    # hide x-axis
    ax.get_xaxis().set_visible(False)
    # ax.title.set_visible(False)
    ax.set_xticklabels(())
    plt.subplots_adjust(hspace=0.001)
    ax.margins(y=0)
    # hide y-axis
    ax.get_yaxis().set_visible(False)
    ax.set_title(title)


def thumbnail_debug(clip):
    for track in clip.tracks:
        (
            stats,
            max_mass,
            max_median_diff,
            min_median_diff,
            max_contour,
        ) = get_track_thumb_stats(clip, track)

        mass_sorted = sorted(
            stats,
            key=lambda s: s.region.mass,
            reverse=True,
        )
        mass_frames = []
        for s in mass_sorted[:10]:
            frame = clip.frame_buffer.get_frame(s.region.frame_number)
            mass_frames.append((frame, s))

        contour_frames = []
        contour_sorted = sorted(
            stats,
            key=lambda s: s.contours,
            reverse=True,
        )
        for s in contour_sorted[:10]:
            frame = clip.frame_buffer.get_frame(s.region.frame_number)
            contour_frames.append((frame, s))

        median_diff = []
        median_sorted = sorted(
            stats,
            key=lambda s: s.median_diff,
            reverse=True,
        )
        for s in median_sorted[:10]:
            frame = clip.frame_buffer.get_frame(s.region.frame_number)
            median_diff.append((frame, s))

        score_sorted = sorted(
            stats,
            key=lambda s: score(
                s, max_mass, max_median_diff, min_median_diff, max_contour
            ),
            reverse=True,
        )
        scored_frames = []

        for s in score_sorted[:10]:
            frame = clip.frame_buffer.get_frame(s.region.frame_number)
            scored_frames.append((frame, s))

        h_info = [mass_frames, contour_frames, median_diff, scored_frames]
        display_track(
            h_info,
            f"{clip.get_id()}-{track.start_frame}-{track.end_frame}",
        )



from abc import ABC, abstractmethod
import logging
import time
import math
import cv2
import numpy as np

from track.track import Track
from track.region import Region
from ml_tools.rectangle import Rectangle
from ml_tools.imageprocessing import normalize, hist_diff


class ClipTracker(ABC):
    def __init__(
        self,
        config,
        cache_to_disk=False,
        keep_frames=True,
        calc_stats=True,
        verbose=False,
        do_tracking=True,
        scale=None,
        calculate_thumbnail_info=False,
    ):
        config = config.get(self.type)
        self.scale = scale
        self.calculate_thumbnail_info = calculate_thumbnail_info
        # if scale:
        # config.rescale(scale)
        self.do_tracking = do_tracking
        self.verbose = verbose
        self.config = config
        self.stats = None
        self.cache_to_disk = cache_to_disk
        self.max_tracks = config.max_tracks
        # frame_padding < 3 causes problems when we get small areas...
        self.frame_padding = max(3, self.config.frame_padding)
        # the dilation effectively also pads the frame so take it into consideration.
        # self.frame_padding = max(0, self.frame_padding - self.config.dilation_pixels)
        self.keep_frames = keep_frames
        self.calc_stats = calc_stats
        self._tracking_time = None
        self.min_dimension = config.min_dimension
        self.background_alg = None
        # if self.config.dilation_pixels > 0:
        #     size = self.config.dilation_pixels * 2 + 1
        #     self.dilate_kernel = np.ones((size, size), np.uint8)

    @abstractmethod
    def type(self):
        """Tracker type IR or Thermal"""

    @abstractmethod
    def parse_clip(self, clip, process_background=False):
        """parse_clip version"""

    @property
    @abstractmethod
    def tracker_version(self):
        """Tracker version"""
        ...

    @property
    @abstractmethod
    def tracking_time(self):
        """Tracker time"""
        ...

    @property
    @abstractmethod
    def start_tracking(
        self, clip, preview_frames, track_frames=True, background_alg=None
    ):
        """start_tracking"""
        ...

    @abstractmethod
    def process_frame(self, clip, rawframe, ffc_affected=False):
        """Get all regions for this sample"""
        ...

    def apply_track_filtering(self, clip):
        self.filter_tracks(clip)
        # apply smoothing if required
        if self.config.track_smoothing and clip.current_frame > 0:
            for track in clip.active_tracks:
                track.smooth(Rectangle(0, 0, clip.res_x, clip.res_y))

    def _get_filtered_frame(self, clip, thermal, sub_change=True, denoise=True):
        """
        Calculates filtered frame from thermal
        :param thermal: the thermal frame
        :param background: (optional) used for background subtraction
        :return: uint8 filtered frame and adjusted clip threshold for normalized frame
        """

        filtered = np.float32(thermal.copy())
        if sub_change:
            avg_change = int(
                round(np.average(thermal) - self.background_alg.get_average())
            )
        else:
            avg_change = 0

        np.clip(
            filtered - self.background_alg.background - avg_change,
            0,
            None,
            out=filtered,
        )
        filtered, stats = normalize(filtered, new_max=255)
        if denoise:
            filtered = cv2.fastNlMeansDenoising(np.uint8(filtered), None)
        if stats[1] == stats[2]:
            mapped_thresh = clip.background_thresh
        else:
            mapped_thresh = clip.background_thresh / (stats[1] - stats[2]) * 255
        return filtered, mapped_thresh

    def _apply_region_matchings(self, clip, regions):
        """
        Work out the best matchings between tracks and regions of interest for the current frame.
        Create any new tracks required.
        """
        unmatched_regions, matched_tracks = self._match_existing_tracks(clip, regions)
        new_tracks = self._create_new_tracks(clip, unmatched_regions)

        unactive_tracks = clip.active_tracks - matched_tracks - new_tracks
        clip.active_tracks = matched_tracks | new_tracks
        self._filter_inactive_tracks(clip, unactive_tracks)

    def _match_existing_tracks(self, clip, regions):
        scores = []
        used_regions = set()
        unmatched_regions = set(regions)
        active = list(clip.active_tracks)
        active.sort(key=lambda x: x.get_id())
        for track in active:
            scores.extend(track.match(regions))

        # makes tracking consistent by ordering by score then by frame since target then track id
        scores.sort(
            key=lambda record: record[1].frames_since_target_seen
            + float(".{}".format(record[1]._id))
        )
        scores.sort(key=lambda record: record[0])
        matched_tracks = set()
        blanked_tracks = set()
        cur_frame = clip.frame_buffer.current_frame
        for score, track, region in scores:
            if (
                track in matched_tracks
                or region in used_regions
                or track in blanked_tracks
            ):
                continue
            logging.debug(
                "frame# %s matched %s to track %s", clip.current_frame, region, track
            )
            used_regions.add(region)
            unmatched_regions.remove(region)
            if not self.config.filter_regions_pre_match:
                if self.config.min_hist_diff is not None:
                    background = self.background_alg.background
                    # if self.scale:
                    #     background = clip.rescaled_background(
                    #         (int(self.res_x), int(self.res_y))
                    #     )
                    hist_v = hist_diff(region, background, cur_frame.thermal)
                    if hist_v > self.config.min_hist_diff:
                        logging.warn(
                            "%s filtering region %s because of hist diff %s track %s ",
                            region.frame_number,
                            region,
                            hist_v,
                            track,
                        )

                        blanked_tracks.add(track)
                        continue
                if (
                    region.pixel_variance < self.config.aoi_pixel_variance
                    or region.mass < self.config.aoi_min_mass
                ):
                    # this will force a blank frame to be added, rather than if we filter earlier
                    # and match this track to a different region
                    logging.debug(
                        "%s filtering region %s because of variance %s and mass %s track %s",
                        region.frame_number,
                        region,
                        region.pixel_variance,
                        region.mass,
                        track,
                    )
                    blanked_tracks.add(track)
                    continue
            track.add_region(region)
            matched_tracks.add(track)

        return unmatched_regions, matched_tracks

    def _create_new_tracks(self, clip, unmatched_regions):
        """Create new tracks for any unmatched regions"""
        new_tracks = set()
        for region in unmatched_regions:
            # make sure we don't overlap with existing tracks.  This can happen if a tail gets tracked as a new object
            overlaps = [
                track.last_bound.overlap_area(region) for track in clip.active_tracks
            ]
            if len(overlaps) > 0 and max(overlaps) > (region.area * 0.25):
                continue

            track = Track.from_region(
                clip,
                region,
                self.tracker_version,
                tracking_config=self.config,
            )
            new_tracks.add(track)
            clip._add_active_track(track)
            self.print_if_verbose(
                "Creating a new track {} with region {} mass{} area {} frame {}".format(
                    track.get_id(),
                    region,
                    track.last_bound.mass,
                    track.last_bound.area,
                    region.frame_number,
                )
            )
        return new_tracks

    def _filter_inactive_tracks(self, clip, unactive_tracks):
        """Filters tracks which are or have become inactive"""
        for track in unactive_tracks:
            track.add_blank_frame()
            if track.tracking:
                clip.active_tracks.add(track)
                logging.debug(
                    "frame {} adding a blank frame to {} ".format(
                        clip.current_frame, track.get_id()
                    )
                )

    def get_delta_frame(self, clip):
        frame = clip.frame_buffer.current_frame
        prev_frame = clip.frame_buffer.prev_frame
        if prev_frame is None:
            return None, None
        filtered, _ = normalize(frame.filtered, new_max=255)
        prev_filtered, _ = normalize(prev_frame.filtered, new_max=255)
        delta_filtered = np.abs(np.float32(filtered) - np.float32(prev_filtered))

        thermal, _ = normalize(frame.thermal, new_max=255)
        prev_thermal, _ = normalize(prev_frame.thermal, new_max=255)
        delta_thermal = np.abs(np.float32(thermal) - np.float32(prev_thermal))
        return delta_thermal, delta_filtered

    def _get_regions_of_interest(self, clip, component_details, centroids=None):
        """
        Calculates pixels of interest mask from filtered image, and returns both the labeled mask and their bounding
        rectangles.
        :param filtered: The filtered frame
        :return: regions of interest, mask frame
        """
        delta_thermal, delta_filtered = self.get_delta_frame(
            clip,
        )

        # we enlarge the rects a bit, partly because we eroded them previously, and partly because we want some context.
        padding = self.frame_padding
        # find regions of interest
        regions = []
        for i, component in enumerate(component_details):
            if centroids is None:
                centroid = [
                    int(component[0] + component[2] / 2),
                    int(component[1] + component[3] / 2),
                ]
            else:
                centroid = centroids[i]
            region = Region(
                component[0],
                component[1],
                component[2],
                component[3],
                mass=component[4],
                id=i,
                frame_number=clip.current_frame,
                centroid=centroid,
            )

            if self.scale:
                region.rescale(1 / self.scale)

            # Make this min area
            if region.width < self.min_dimension or region.height < self.min_dimension:
                continue
            # GP this needs to be checked for themals 29/06/2022
            if clip.type == "IR":
                if delta_thermal is not None:
                    # filtered only 0 or 255
                    sub_delta = region.subimage(delta_thermal)
                    previous_delta_mass = len(
                        sub_delta[sub_delta > clip.background_thresh]
                    )
                    # if previous_delta_mass == 0:
                    #     logging.info("No mass from previous so skipping")
                    #     continue
                    region.pixel_variance = np.var(sub_delta)

            elif delta_filtered is not None:
                region_difference = region.subimage(delta_filtered)
                region.pixel_variance = np.var(region_difference)
            old_region = region.copy()
            region.crop(clip.crop_rectangle)
            region.was_cropped = str(old_region) != str(region)

            if self.config.cropped_regions_strategy == "cautious":
                crop_width_fraction = (
                    old_region.width - region.width
                ) / old_region.width
                crop_height_fraction = (
                    old_region.height - region.height
                ) / old_region.height
                if crop_width_fraction > 0.25 or crop_height_fraction > 0.25:
                    continue
            elif (
                self.config.cropped_regions_strategy == "none"
                or self.config.cropped_regions_strategy is None
            ):
                if region.was_cropped:
                    continue
            elif self.config.cropped_regions_strategy != "all":
                raise ValueError(
                    "Invalid mode for CROPPED_REGIONS_STRATEGY, expected ['all','cautious','none'] but found {}".format(
                        self.config.cropped_regions_strategy
                    )
                )

            # filter out regions that are probably just noise
            if self.config.filter_regions_pre_match and (
                region.pixel_variance < self.config.aoi_pixel_variance
                and region.mass < self.config.aoi_min_mass
            ):
                logging.debug(
                    "%s filtering region %s because of variance %s and mass %s",
                    region.frame_number,
                    region,
                    region.pixel_variance,
                    region.mass,
                )
                continue

            region.enlarge(padding, max=clip.crop_rectangle)
            # gp dunno if we should use this feels like we already have the edge
            # extra_edge = 0
            extra_edge = math.ceil(clip.crop_rectangle.width * 0.03)
            region.set_is_along_border(clip.crop_rectangle, edge=extra_edge)
            regions.append(region)
        return regions

    def filter_tracks(self, clip):
        for track in clip.tracks:
            track.trim()
            track.set_end_s(clip.frames_per_second)

        [track.calculate_stats() for track in clip.tracks]
        clip.tracks.sort(reverse=True, key=lambda record: record.stats.score)
        if self.verbose:
            for track in clip.tracks:
                start_s, end_s = clip.start_and_end_in_secs(track)
                logging.info(
                    " - track %s duration: %.1fsec, number of frames:%s, stats %s",
                    track.get_id(),
                    end_s - start_s,
                    len(track),
                    track.stats,
                )
        # filter out tracks that probably are just noise.
        good_tracks = []
        self.print_if_verbose(
            "{} {}".format("Number of tracks before filtering", len(clip.tracks))
        )

        for track in clip.tracks:
            # discard any tracks that overlap too often with other tracks.  This normally means we are tracking the
            # tail of an animal.
            if not self.filter_track(clip, track):
                good_tracks.append(track)

        clip.tracks = good_tracks
        self.print_if_verbose(
            "{} {}".format("Number of 'good' tracks", len(clip.tracks))
        )
        # apply max_tracks filter
        # note, we take the n best tracks.
        if self.max_tracks is not None and self.max_tracks < len(clip.tracks):
            logging.warning(
                " -using only {0} tracks out of {1}".format(
                    self.max_tracks, len(clip.tracks)
                )
            )
            clip.filtered_tracks.extend(
                [("Too many tracks", track) for track in clip.tracks[self.max_tracks :]]
            )
            clip.tracks = clip.tracks[: self.max_tracks]

        for key in clip.filtered_tracks:
            self.print_if_verbose(
                "filtered track {} because {}".format(key[1].get_id(), key[0])
            )

    def filter_track(self, clip, track):
        stats = track.stats
        # discard any tracks that are less min_duration
        # these are probably glitches anyway, or don't contain enough information.
        if len(track) < self.config.min_duration_secs * clip.frames_per_second:
            self.print_if_verbose("Track filtered. Too short, {}".format(len(track)))
            clip.filtered_tracks.append(("Track filtered.  Too short", track))
            return True

        # discard tracks that do not move enough

        if (
            stats.max_offset < self.config.track_min_offset
            or stats.frames_moved < self.config.min_moving_frames
        ):
            self.print_if_verbose(
                "Track filtered.  Didn't move {}".format(stats.max_offset)
            )
            clip.filtered_tracks.append(("Track filtered.  Didn't move", track))
            return True

        if stats.blank_percent > self.config.max_blank_percent:
            self.print_if_verbose("Track filtered.  Too Many Blanks")
            clip.filtered_tracks.append(("Track filtered. Too Many Blanks", track))
            return True
        if stats.region_jitter > self.config.max_jitter:
            self.print_if_verbose("Track filtered.  Too Jittery")
            clip.filtered_tracks.append(("Track filtered.  Too Jittery", track))
            return True
        # discard tracks that do not have enough delta within the window (i.e. pixels that change a lot)
        if stats.delta_std < clip.track_min_delta:
            self.print_if_verbose(
                "Track filtered.  Too static {}".format(stats.delta_std)
            )
            clip.filtered_tracks.append(("Track filtered.  Too static", track))
            return True
        if stats.delta_std > clip.track_max_delta:
            self.print_if_verbose("Track filtered.  Too Dynamic")
            clip.filtered_tracks.append(("Track filtered.  Too Dynamic", track))
            return True
        # discard tracks that do not have enough enough average mass.
        if stats.average_mass < self.config.track_min_mass:
            self.print_if_verbose(
                "Track filtered.  Mass too small ({})".format(stats.average_mass)
            )
            clip.filtered_tracks.append(("Track filtered.  Mass too small", track))

            return True
        # GP Removed as dont think this does much but filter good tracks, if the tracking
        # chose 2 tracks lets keep them, perhaps usefull for short tracks where 90% overlaps
        # but these will probably be filtered by being too short
        # highest_ratio = 0
        # for other in clip.tracks:
        #     if track == other:
        #         continue
        #     highest_ratio = max(track.get_overlap_ratio(other), highest_ratio)
        #
        # if highest_ratio > self.config.track_overlap_ratio:
        #     self.print_if_verbose(
        #         "Track filtered.  Too much overlap {}".format(highest_ratio)
        #     )
        #     clip.filtered_tracks.append(("Track filtered.  Too much overlap", track))
        #     return True

        return False

    def print_if_verbose(self, info_string):
        if self.verbose:
            logging.info(info_string)


class Background(ABC):
    TRIGGER_FRAMES = 2

    def __init__(self):
        self.rescaled = None
        self.prev_triggered = False
        self.triggered = 0
        self.movement_detected = False
        self.kernel_trigger = np.ones(
            (15, 15), "uint8"
        )  # kernel for erosion when not recording
        self.kernel_recording = np.ones(
            (10, 10), "uint8"
        )  # kernel for erosion when recording

    @abstractmethod
    def set_background(self, background, frames=1):
        """set_background version"""
        ...

    @abstractmethod
    def update_background(self, thermal, filtered):
        """update_background version"""
        ...

    @abstractmethod
    def compute_filtered(self, thermal, threshold):
        """compute_filtered version"""
        ...

    @property
    @abstractmethod
    def background(self):
        """Background image"""
        ...

    @property
    @abstractmethod
    def frames(self):
        """frames used"""
        ...

    @property
    def frames(self):
        return self._frames

    def get_kernel(self):
        if self.movement_detected:
            return self.kernel_recording
        else:
            return self.kernel_trigger

    def detect_motion(self):
        fg = self.compute_filtered(None)
        erosion_image = cv2.erode(fg, self.get_kernel())
        erosion_pixels = len(erosion_image[erosion_image > 0])

        self.prev_triggered = erosion_pixels > 0
        if erosion_pixels > 0:
            self.triggered += 1
            self.triggered = min(self.triggered, 2)
        else:
            self.triggered -= 1
            self.triggered = max(self.triggered, 0)
        self.movement_detected = self.triggered >= Background.TRIGGER_FRAMES
        return self.movement_detected


class CVBackground(Background):
    def __init__(self, tracking_alg="mog2"):
        super().__init__()
        self.use_subsense = False
        # knn doesnt respect learning rate, but maybe mog2 is better anyway
        if tracking_alg == "subsense":
            import pybgs as bgs

            self.use_subsense = True

            self.algorithm = bgs.SuBSENSE()
        elif tracking_alg == "mog2":
            self.algorithm = cv2.createBackgroundSubtractorMOG2(
                history=1000, detectShadows=False
            )
        else:
            raise Exception(f"No algorihtm details found for {tracking_alg}")
            # print(self.algorithm.getBackgroundRatio(), "RATION")
            # 1 / 0
        # self.algorithm = cv2.createBackgroundSubtractorKNN(
        #     history=1000, detectShadows=False
        # )
        self._frames = 0
        self._background = None

    def set_background(self, background, frames=1):
        # seems to be better to do x times rather than just set background
        if self.use_subsense:
            for _ in range(10):
                # doesnt have a learning rate
                self.update_background(background, learning_rate=1)
        else:
            self.update_background(background, learning_rate=1)

            # return

    def update_background(self, thermal, filtered=None, learning_rate=-1):
        if self.use_subsense:
            self._background = self.algorithm.apply(thermal)
        else:
            self._background = self.algorithm.apply(thermal, None, learning_rate)

        self._frames += 1

    @property
    def background(self):
        if self.use_subsense:
            return self.algorithm.getBackgroundModel()
        else:
            return self.algorithm.getBackgroundImage()

    def compute_filtered(self, thermal):
        return self._background


class DiffBackground(Background):
    def __init__(self, background_thresh):
        super().__init__()
        self._frames = 1
        self._background = None
        self.background_thresh = background_thresh

    def set_background(self, background, frames=1):
        self._frames = frames
        self._background = np.float32(background) * self.frames
        return

    def update_background(self, thermal):
        background = self.background
        filtered = get_diff_back_filtered(
            self.background,
            thermal,
            self.background_thresh,
        )
        new_thermal = np.where(filtered > 0, background, thermal)
        self._background += new_thermal
        self._frames += 1

    def compute_filtered(self, thermal=None):
        filtered = get_diff_back_filtered(
            self.background,
            thermal,
            self.background_thresh,
        )
        return filtered

    @property
    def background(self):
        return self._background / self.frames

    @property
    def frames(self):
        return self._frames


def get_diff_back_filtered(background, frame, back_thresh):
    """
    Calculates filtered frame from thermal
    :param frame: the frame
    :param background: (optional) used for background subtraction
    :return: uint8 filtered frame and adjusted clip threshold for normalized frame
    """

    filtered = np.float32(frame.copy())
    filtered = abs(filtered - background)
    filtered[filtered < back_thresh] = 0
    filtered, stats = normalize(filtered, new_max=255)
    return filtered



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os
import logging
import numpy as np
import time
import yaml
from datetime import datetime

import cv2

from ml_tools.tools import eucl_distance_sq
from track.region import Region
from ml_tools.imageprocessing import (
    detect_objects_ir,
)
from track.cliptracker import ClipTracker, CVBackground

DO_SALIENCY = False
DEBUG_TRAP = False


class Line:
    def __init__(self, m, c):
        self.m = m
        self.c = c

    # check point is above live
    def is_above(self, point):
        y_value = self.y_res(point[0])
        if point[1] > y_value:
            return True
        return False

    # check point is below line
    def is_below(self, point):
        return not self.is_above(point)

    # check point is to the left of line
    def is_left(self, point):
        x_value = self.x_res(point[1])
        if point[0] < x_value:
            return True
        return False

    # check point is to the right of line
    def is_right(self, point):
        return not self.is_left(point)

    def y_res(self, x):
        return x * self.m + self.c

    def x_res(self, y):
        return (y - self.c) / self.m
        # return x * self.m + self.c

    def __str__(self):
        return f"y={self.m}x + {self.c}"


def get_trap_lines(trap_size):
    if trap_size == "S":
        lb = Line(1.3, 297.5)
        rb = Line(-1.4, 1148)
    else:
        # just default to large for now
        lb = Line(1.28, 180)
        rb = Line(-1.2, 979)
    logging.info(
        "Getting trap lines for trap size %s left bottom %s right bottom %s",
        trap_size,
        lb,
        rb,
    )
    return lb, rb


class IRTrackExtractor(ClipTracker):
    PREVIEW = "preview"
    VERSION = 10
    TYPE = "IR"
    # # new perspective 19-10-2022
    # LEFT_BOTTOM = Line(1.28, 180)
    # RIGHT_BOTTOM = Line(-1.2, 979)
    #
    # LEFT_BOTTOM = Line(1.3, 297.5)
    #
    # RIGHT_BOTTOM = Line(-1.4, 1148)
    # old perspective
    # LEFT_BOTTOM = Line(5 / 14, 160)
    # RIGHT_BOTTOM = Line(-5 / 12, 421.7)

    @property
    def type(self):
        return IRTrackExtractor.TYPE

    @property
    def tracker_version(self):
        return self.version

    @property
    def tracking_time(self):
        return self._tracking_time

    def __init__(
        self,
        config,
        cache_to_disk=False,
        keep_frames=True,
        calc_stats=True,
        verbose=False,
        scale=None,
        do_tracking=True,
        on_trapped=None,
        update_background=True,
        trap_size="L",
        tracking_alg="mog2",
        check_trapped=False,
        from_pi=False,
    ):
        super().__init__(
            config,
            cache_to_disk,
            keep_frames,
            calc_stats,
            verbose,
            do_tracking=do_tracking,
            scale=scale,
        )

        if from_pi:
            self.version = f"PI-IR-{IRTrackExtractor.VERSION}"
        else:
            self.version = f"IR-{IRTrackExtractor.VERSION}"

        self.check_trapped = check_trapped
        self.tracking_alg = tracking_alg
        self.on_trapped = on_trapped
        self.saliency = None
        self.background = None
        self.res_x = None
        self.res_y = None
        # if false background is being updated else where i.e motion detector
        self.update_background = update_background
        self.trap_size = trap_size
        self.left_bottom, self.right_bottom = get_trap_lines(self.trap_size)
        self.learning_rate = -1

    def parse_clip(self, clip, process_background=False):
        """
        Loads a cptv file, and prepares for track extraction.
        """
        clip.type = self.type
        self._tracking_time = None
        start = time.time()
        clip.set_frame_buffer(
            False,
            self.cache_to_disk,
            False,
            self.keep_frames,
            max_frames=(
                None if self.keep_frames else 51
            ),  # enough to cover back comparison
        )

        _, ext = os.path.splitext(clip.source_file)
        background = None
        vidcap = cv2.VideoCapture(str(clip.source_file))
        fail_count = 0

        while True:
            success, image = vidcap.read()

            if not success:
                if fail_count < 1:
                    fail_count += 1
                    # try once more if its first fail as the mp4s from pi have errors at key frames
                    continue
                break
            fail_count = 0

            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            if background is None:
                self.res_x = gray.shape[0]
                self.res_y = gray.shape[1]
                clip.set_res(gray.shape[1], gray.shape[0])
                if clip.from_metadata:
                    for track in clip.tracks:
                        track.crop_regions()

                background = np.uint8(gray)
                # cv2.imshow("bak", np.uint8(background))
                # cv2.waitKey(1000)
                # if this is grey scale it is the tracking backgorund frame
                is_background_frame = np.all(
                    image[:, :, 0] == image[:, :, 1]
                ) and np.all(image[:, :, 1] == image[:, :, 2])
                background_frames = 500 if is_background_frame else 1
                self.start_tracking(
                    clip, background_frame=gray, background_frames=background_frames
                )
                if is_background_frame:
                    continue
            self.process_frame(clip, gray)

        vidcap.release()
        if not clip.from_metadata and self.do_tracking:
            self.apply_track_filtering(clip)

        if self.calc_stats:
            clip.stats.completed()

        self._tracking_time = time.time() - start
        return True

    def start_tracking(
        self,
        clip,
        frames=None,
        track_frames=-1,
        background_alg=None,
        background_frame=None,
        background_frames=1,
        retrack_back=True,
    ):
        # track frames is the number of frames to track from supplied frames (most recent first)
        # -1 means all, 0 means none, 1 means just last frame etc
        self.res_x = clip.res_x
        self.res_y = clip.res_y
        if DO_SALIENCY:
            self.init_saliency()
        clip.set_model("IR")
        clip.set_video_stats(datetime.now())
        if background_alg is None:
            self.background = CVBackground(self.tracking_alg)
            # self.diff_background = DiffBackground(clip.background_thresh)
            if background_frame is not None:
                # if self.scale:
                #     logging.info("Rescaling back")
                #     background_frame = cv2.resize(
                #         background_frame,
                #         (int(self.res_x * self.scale), int(self.res_y * self.scale)),
                #         interpolation=cv2.INTER_AREA,
                #     )

                self.background.set_background(background_frame, background_frames)
                # self.diff_background.set_background(background, background_frames)
        else:
            self.background = background_alg
        self.init_saliency()
        if frames is not None:
            do_tracking = self.do_tracking
            update_background = self.update_background
            remaining = len(frames)
            for i, frame in enumerate(frames):
                self.do_tracking = do_tracking and (
                    (track_frames == -1) or (remaining <= track_frames)
                )
                self.learning_rate = 0
                self.update_background = self.do_tracking and retrack_back
                self.process_frame(clip, frame)
                remaining -= 1
            self.learning_rate = -1
            self.update_background = update_background
            self.do_tracking = do_tracking

    def init_saliency(self):
        res_x = self.res_x
        res_y = self.res_y
        if self.scale is not None:
            res_x = int(self.res_x * self.scale)
            res_y = int(self.res_y * self.scale)
        if DO_SALIENCY:
            self.saliency = cv2.saliency.MotionSaliencyBinWangApr2014_create()
            self.saliency.setImagesize(res_x, res_y)
            self.saliency.init()

    def process_frame(self, clip, frame, ffc_affected=False):
        start = time.time()
        if len(frame.shape) == 3:
            # in rgb so convert to gray
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        if self.saliency is None and DO_SALIENCY:
            if self.resize_dims is not None:
                self.init_saliency(
                    int(self.res_x * self.scale), int(self.res_y * self.scale)
                )
            else:
                self.init_saliency(clip.res_x, clip.res_y)

        if ffc_affected:
            self.print_if_verbose("{} ffc_affected".format(clip.current_frame))
        clip.ffc_affected = ffc_affected
        self._process_frame(clip, frame, ffc_affected)

    def _get_filtered_frame_ir(self, thermal, repeats=1):
        if not DO_SALIENCY:
            return thermal, 0
        for _ in range(repeats):
            (success, saliencyMap) = self.saliency.computeSaliency(thermal)
        saliencyMap = (saliencyMap * 255).astype("uint8")
        return saliencyMap, 0

    # merge all regions that the midpoint is within the max(width,height) from the midpoint of another region
    # keep merging until no more merges are possible, tihs works paticularly well from the IR videos where
    # the filtered image is quite fragmented
    def merge_components(self, rectangles):
        min_mass = 10 * 4
        min_size = 16
        MAX_GAP = 40
        if self.scale:
            min_mass = int(min_mass * self.scale)
            min_size = int(min_size * self.scale)
            MAX_GAP *= self.scale
        rectangles = [
            r
            for r in rectangles
            if r[4] > min_mass or (r[2] > min_size and r[3] > min_size)
        ]
        rectangles = sorted(rectangles, key=lambda s: s[4], reverse=True)
        # only merge on original rectangle not on resized
        for i in range(len(rectangles)):
            rectangles[i] = (rectangles[i], rectangles[i].copy())
        rect_i = 0
        rectangles = list(rectangles)
        while rect_i < len(rectangles):
            rect, merged_r = rectangles[rect_i]
            merged = False
            mid_x = rect[2] / 2.0 + rect[0]
            mid_y = rect[3] / 2.0 + rect[1]
            index = 0
            while index < len(rectangles):
                within = False
                r_2 = rectangles[index][0]
                if r_2[0] == rect[0]:
                    index += 1
                    continue

                r_mid_x = r_2[2] / 2.0 + r_2[0]
                r_mid_y = r_2[3] / 2.0 + r_2[1]
                distance = (mid_x - r_mid_x) ** 2 + (r_mid_y - mid_y) ** 2
                distance = distance**0.5

                if r_2[2] + rect[2] > max(r_2[0] + r_2[2], rect[2] + rect[0]) - min(
                    r_2[0], rect[0]
                ):
                    within = r_2[3] + rect[3] > max(
                        r_2[1] + r_2[3], rect[1] + rect[3]
                    ) - min(r_2[1], rect[1])
                distance = rect_distance(rect, r_2)
                if distance < MAX_GAP or within:
                    cur_right = merged_r[0] + merged_r[2]
                    cur_bottom = merged_r[0] + merged_r[2]

                    merged_r[0] = min(merged_r[0], r_2[0])
                    merged_r[1] = min(merged_r[1], r_2[1])
                    merged_r[2] = max(cur_right, r_2[0] + r_2[2])
                    merged_r[3] = max(merged_r[1] + merged_r[3], r_2[1] + r_2[3])
                    merged_r[2] -= merged_r[0]
                    merged_r[3] -= merged_r[1]
                    merged_r[4] += r_2[4]
                    merged = True
                    # break
                    del rectangles[index]
                else:
                    index += 1
            if merged:
                rect_i = 0
            else:
                rect_i += 1
        rectangles = [rect[1] for rect in rectangles]
        return rectangles

    def _process_frame(self, clip, frame, ffc_affected=False):
        wait = 1
        """
        Tracks objects through frame
        :param thermal: A numpy array of shape (height, width) and type uint16
            If specified background subtraction algorithm will be used.
        """

        saliencyMap = None
        filtered = None
        if self.do_tracking:
            if self.background._background is None:
                self.background.set_background(frame.copy())
            # GP NOT using probably can remove
            if DO_SALIENCY:
                repeats = 1
                if clip.current_frame < 6:
                    # helps init the saliency, when ir tracks have a 5 second preview this shouldnt be needed
                    repeats = 6
                saliency_thermal = None
                tracking_thermal = frame
                if self.scale:
                    tracking_thermal = cv2.resize(
                        tracking_thermal,
                        (int(self.res_x * self.scale), int(self.res_y * self.scale)),
                        interpolation=cv2.INTER_AREA,
                    )
                saliencyMap, _ = self._get_filtered_frame_ir(
                    tracking_thermal, repeats=repeats
                )
            if self.update_background:
                has_motion = self.background.movement_detected
                learning_rate = 0 if has_motion else self.learning_rate
                self.background.update_background(
                    frame, learning_rate=self.learning_rate
                )

            filtered = self.background.compute_filtered(frame)

            if not clip.background_calculated:
                clip.set_background(self.background.background)

        threshold = 0
        if np.amin(saliencyMap) == 255:
            num = 0
            mask = saliencyMap.copy()
            component_details = []
            saliencyMap[:] = 0
        else:
            pass
            # backsub = np.where(saliencyMap > 0, saliencyMap, backsub)
        cur_frame = clip.add_frame(frame, filtered, saliencyMap, ffc_affected)

        if not self.do_tracking:
            return
        re_f = filtered
        if self.scale:
            re_f = cv2.resize(
                filtered,
                (int(self.res_x * self.scale), int(self.res_y * self.scale)),
                interpolation=cv2.INTER_AREA,
            )
        num, mask, component_details = detect_objects_ir(re_f, threshold=0)
        component_details = component_details[1:]
        component_details = self.merge_components(component_details)
        if clip.from_metadata:
            for track in clip.tracks:
                if clip.current_frame in track.frame_list:
                    track.add_frame_for_existing_region(
                        cur_frame,
                        threshold,
                        (
                            clip.frame_buffer.prev_frame.filtered
                            if clip.frame_buffer.prev_frame is not None
                            else None
                        ),
                    )
        else:
            regions = []
            if ffc_affected:
                clip.active_tracks = set()
            else:
                s = time.time()
                regions = self._get_regions_of_interest(clip, component_details)
                self._apply_region_matchings(clip, regions)
            for track in clip.active_tracks:
                if track.trap_reported:
                    continue
                self.inside_trap_top(track, self.scale)
                if track.in_trap:
                    filtered = self.filter_track(clip, track, track.get_stats())
                    if not filtered:
                        track.trigger_frame = cur_frame.frame_number
                        # fire trapping event
                        if self.on_trapped is not None:
                            track.trap_reported = True
                            self.on_trapped(track)
            clip.region_history.append(regions)
            if DEBUG_TRAP:
                if len(clip.tracks) > 0:
                    self.show_trap_info(clip, frame)
        # self.debug_tracking(frame, filtered, mask, component_details)

    def debug_tracking(self, frame, filtered, mask, components):
        region_copy = filtered.copy()
        for r in components:
            region = Region.from_ltwh(*r[:4])
            region.rescale(4)
            region = region.to_array()
            cv2.rectangle(
                region_copy,
                (region[0], region[1]),
                (region[2], region[3]),
                color=(255, 255, 255),
                thickness=3,
            )
        cv2.imshow("gray", frame)
        cv2.imshow("f", region_copy)
        cv2.imshow("b", self.background.background)
        cv2.moveWindow("gray", 0, 0)
        cv2.moveWindow("f", 640, 0)
        cv2.waitKey()

    def show_trap_info(self, clip, frame):
        image = frame.copy()
        in_trap = False
        wait = 100
        start = (0, 480 - int(self.left_bottom.y_res(0)))
        end = (int(self.left_bottom.x_res(480)), 0)

        image = cv2.line(image, start, end, (255, 255, 0), 10)
        start = (int(self.right_bottom.x_res(480)), 0)
        end = (int(self.right_bottom.x_res(0)), 480)

        image = cv2.line(image, start, end, (255, 255, 0), 10)

        for track in clip.active_tracks:
            filtered = self.filter_track(clip, track, track.get_stats())
            in_trap = in_trap or (track.in_trap and not filtered)
            region = track.bounds_history[-1]
            # if self.scale:
            #     region = region.copy()
            #     region.rescale(1 / self.scale)
            region.in_trap = track.last_bound.in_trap
            start_point = (int(region.left), int(region.top))
            end_point = (int(region.right), int(region.bottom))
            image = cv2.rectangle(image, start_point, end_point, (255, 0, 0), 2)
            if region.in_trap:
                text_point = (start_point[0], start_point[1] + 50)
                image = cv2.putText(
                    image,
                    f"Trapped?{ track.in_trap} f?{filtered} {track.direction}",
                    text_point,
                    cv2.FONT_HERSHEY_SIMPLEX,
                    1,
                    (255, 0, 0),
                )
                # wait = 300
        if in_trap:
            image = cv2.putText(
                image,
                f"CAUGHT EM",
                (50, 50),
                cv2.FONT_HERSHEY_SIMPLEX,
                1,
                (255, 0, 0),
            )
            wait = 300

        cv2.imshow("id", image)
        cv2.waitKey(100)
        # 1 / 0

    def filter_components(self, component_details):
        filtered = []
        for component in component_details:
            region = Region(
                component[0],
                component[1],
                component[2],
                component[3],
            )
            p = (region.right, 480 - region.bottom)

            filter = (
                self.left_bottom.is_above(p)
                and self.left_bottom.is_left(p)
                # or self.right_bottom.is_above(region)
                # or IRTrackExtractor.BACK_BOTTOM.is_above(region)
            )
            p = (region.left, 480 - region.bottom)

            filter = filter or (
                self.right_bottom.is_above(p) and self.right_bottom.is_right(p)
            )
            p = (region.left, 480 - region.bottom)

            # filter = filter or (IRTrackExtractor.BACK_TOP.is_above(p))
            if not filter:
                filtered.append(component)
            else:
                logging.info("Filtered components %s", region)
        return filtered

    def filter_track(self, clip, track, stats):
        # return not track.stable
        # discard any tracks that are less min_duration
        # these are probably glitches anyway, or don't contain enough information.
        if len(track) < self.config.min_duration_secs * clip.frames_per_second:
            self.print_if_verbose("Track filtered. Too short, {}".format(len(track)))
            clip.filtered_tracks.append(("Track filtered.  Too short", track))
            return True

        # discard tracks that do not move enough

        if (
            stats.max_offset < self.config.track_min_offset
            or stats.frames_moved < self.config.min_moving_frames
        ):
            self.print_if_verbose(
                "Track filtered.  Didn't move {}".format(stats.max_offset)
            )
            clip.filtered_tracks.append(("Track filtered.  Didn't move", track))
            return True

        # if stats.blank_percent > self.config.max_blank_percent:
        #     self.print_if_verbose("Track filtered.  Too Many Blanks")
        #     clip.filtered_tracks.append(("Track filtered. Too Many Blanks", track))
        #     return True

        # highest_ratio = 0
        # for other in clip.tracks:
        #     if track == other:
        #         continue
        #     highest_ratio = max(track.get_overlap_ratio(other), highest_ratio)
        #
        # if highest_ratio > self.config.track_overlap_ratio:
        #     self.print_if_verbose(
        #         "Track filtered.  Too much overlap {}".format(highest_ratio)
        #     )
        #     clip.filtered_tracks.append(("Track filtered.  Too much overlap", track))
        #     return True

        return False

    def get_delta_frame(self, clip):
        # GP used to be 50 frames ago
        frame = clip.frame_buffer.current_frame
        frame_ago = 10 if self.keep_frames is None else max(10, self.keep_frames)
        if clip.current_frame < frame_ago:
            prev_i = clip.current_frame - 1
        else:
            prev_i = frame_ago
        s = time.time()
        prev_frame = clip.frame_buffer.get_frame_ago(prev_i)

        if (
            prev_frame is None
            or prev_i == frame.frame_number
            or prev_frame.filtered is None
        ):
            return None, None

        s = time.time()

        delta_filtered = np.abs(frame.filtered - prev_frame.filtered)
        delta_thermal = np.abs(frame.thermal - prev_frame.thermal)
        return delta_thermal, delta_filtered

    def inside_trap_bottom(self, track, scale=None):
        region = track.last_bound.copy()
        # if scale:
        #     region.rescale(1 / scale)
        if region.width < 60 or region.height < 40:
            # dont want small regions
            return False
        if track.direction == 0:
            if region.left < 100:
                track.direction |= Direction.LEFT
            if region.right > (640 - 100):
                track.direction |= Direction.RIGHT
            if region.bottom > (480 - 100):
                track.direction |= Direction.BOTTOM
            if track.direction == 0:
                if region.bottom < 300:
                    track.direction |= Direction.TOP
                else:
                    track.direction = Direction.MIDDLE
        p = (region.left, 480 - region.bottom)
        # bottom left region below left Line
        # bottom right region below right line

        # alternative method is to use
        # top right is below and rigth of left line
        # top left is below and left of right line
        # this fixes problem when animal comes from behind trap towards camera
        # this triggerrs early for bottom areas
        inside = self.left_bottom.is_below(p) and self.left_bottom.is_right(p)
        p = (region.right, 480 - region.bottom)
        inside = inside and (
            self.right_bottom.is_below(p) and self.right_bottom.is_left(p)
        )

        x_diff = p[0] - self.right_bottom.x_res(p[1])
        inside = inside and abs(x_diff) > 150
        track.last_bound.in_trap = inside
        track.update_trapped_state()
        return inside

    def inside_trap_top(self, track, scale=None):
        SIDE_ALLOWANCE = 150
        TOP_ALLOWANCE = 300
        BOTTOM_ALLOWANCE = 100
        region = track.last_bound.copy()
        # if scale:
        # region.rescale(1 / scale)
        if region.width < 60 or region.height < 40:
            # dont want small regions
            return False

        # Set direction track is coming from
        # this should help filter out some FPS
        if track.direction == 0:
            if region.left < SIDE_ALLOWANCE:
                track.direction |= Direction.LEFT
            if region.right > (640 - SIDE_ALLOWANCE):
                track.direction |= Direction.RIGHT
            if region.bottom > (480 - BOTTOM_ALLOWANCE):
                track.direction |= Direction.BOTTOM
            if track.direction == 0:
                if region.bottom < TOP_ALLOWANCE:
                    track.direction |= Direction.TOP
                else:
                    track.direction = Direction.MIDDLE

        # check top right is to the left and below left trap boundary line
        p = (region.right, 480 - region.top)
        inside = self.left_bottom.is_below(p) and self.left_bottom.is_right(p)
        x_pos = self.left_bottom.x_res(p[1])
        x_diff = abs(p[0] - x_pos)
        # amount of animal inside trap from left
        left_percent = x_diff / region.width

        p = (region.left, 480 - region.top)

        # check top left is to the right and below right trap boundary line
        inside = (
            inside and self.right_bottom.is_below(p) and self.right_bottom.is_left(p)
        )

        x_pos = self.right_bottom.x_res(p[1])
        # could try using bottom  rather than region.top here
        x_diff = abs(p[0] - x_pos)
        # amount of animal inside trap from right
        right_percent = x_diff / region.width

        if not inside:
            return False
        in_trap = False

        # atleast half the animal should be inside the trap, sometimes tails etc
        # can be outside trap so best to trigger a bit early than not at all
        if left_percent < 0.5 and right_percent < 0.5:
            return False
        if track.direction & Direction.LEFT and region.left > 40 and left_percent > 0.5:
            in_trap = True

        elif (
            track.direction & Direction.RIGHT
            and region.right < 580
            and right_percent > 0.5
        ):
            in_trap = True

        if track.direction == Direction.TOP and region.bottom > 300:
            in_trap = True
        if track.direction == Direction.BOTTOM and region.bottom < 480 - 50:
            in_trap = True

        if (
            track.direction == Direction.MIDDLE
            and region.left > 40
            and region.right < 580
        ):
            in_trap = True
        track.last_bound.in_trap = in_trap
        track.update_trapped_state()
        return in_trap


class Direction:
    LEFT = 1
    BOTTOM = 2
    RIGHT = 4
    TOP = 8
    MIDDLE = 16


def rect_distance(r_a, r_b):
    x_1 = 0
    x_2 = 0
    y_1 = 0
    y_2 = 0
    if r_a[2] + r_b[2] > max(r_a[0] + r_a[2], r_b[2] + r_b[0]) - min(r_a[0], r_b[0]):
        # if over lap on  dimension then no distance
        pass
    elif r_a[0] < r_b[0]:
        # right side of r_a to left side r_b
        x_1 = r_a[0] + r_a[2]
        x_2 = r_b[0]
    else:
        # right side of r_b to left side r_a
        x_1 = r_b[0] + r_b[2]
        x_2 = r_a[0]
    overlap = r_a[3] + r_b[3] > max(r_a[1] + r_a[3], r_b[1] + r_b[3]) - min(
        r_a[1], r_b[1]
    )
    if overlap:
        # if over lap on  dimension then no distance
        pass
    elif r_a[1] < r_b[1]:
        # r_a above r_b bottom r_a top of r_b
        y_1 = r_a[1] + r_a[3]
        y_2 = r_b[1]
    else:
        y_1 = r_b[1] + r_b[3]
        y_2 = r_a[1]
    return eucl_distance_sq((x_1, y_1), (x_2, y_2)) ** 0.5



import time
import dbus
import logging
import json

DBUS_NAME = "org.cacophony.TrapController"
DBUS_PATH = "/org/cacophony/TrapController"


def trigger_trap(tag=None):
    bus = dbus.SystemBus()
    try:
        proxy = bus.get_object(DBUS_NAME, DBUS_PATH)
        data = {}
        if tag is not None:
            data["tag"] = tag
        proxy.TriggerTrap(json.dumps(data))
    except:
        logging.error("trapped dbus error ", exc_info=True)



import threading
import socket
from config.config import Config
from ml_tools.logs import init_logging
from classify.clipclassifier import ClipClassifier
import logging
import os
import argparse
import json
import traceback
from ml_tools.tools import CustomJSONEncoder


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("-c", "--config-file", help="Path to config file to use")
    return parser.parse_args()


# Links to socket and continuously waits for 1 connection
def main():
    init_logging()
    args = parse_args()

    config = Config.load_from_file(args.config_file)
    service = ClassifyService(config)
    try:
        service.run()
    except KeyboardInterrupt:
        logging.info("Keyboard interupt closing down")
    except PermissionError:
        logging.error("Error with permissions", exc_info=True)
    except:
        logging.error("Error with service restarting", exc_info=True)


class ClassifyService:
    def __init__(self, config):
        self.config = config

        self.clip_classifier = ClipClassifier(
            config,
        )

    def run(self):
        logging.info("Running on %s", self.config.classify.service_socket)
        max_jobs = 2
        try:
            os.unlink(self.config.classify.service_socket)
        except OSError:
            if os.path.exists(self.config.classify.service_socket):
                raise

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(self.config.classify.service_socket)
        sock.listen(1)
        self.clip_classifier.models = {}
        self.clip_classifier.load_models()

        while True:
            logging.info("waiting for jobs")
            connection, client_address = sock.accept()
            t = threading.Thread(
                target=classify_job,
                args=(self.clip_classifier, connection, client_address),
            )
            t.start()


def classify_job(clip_classifier, clientsocket, addr):
    try:
        job = read_all(clientsocket).decode()
        args = json.loads(job)
        if "file" not in args:
            logging.error("File name must be specified in argument dictionary")
            clientsocket.sendall(
                json.dumps(
                    {"error": "File name must be specified in argument dictionary"}
                ).encode()
            )
            return
        logging.info("Classifying %s with args %s", args["file"], args)

        meta_data = clip_classifier.process_file(
            args["file"],
            cache=args.get("cache"),
            reuse_frames=args.get("reuse_frames"),
        )

        response = clientsocket.sendall(
            json.dumps(meta_data, cls=CustomJSONEncoder).encode()
        )
        if response:
            logging.error("Error sending data to socket %s", response)
    except BrokenPipeError:
        logging.error(
            "Error sending metadata for job %s too %s",
            args["file"],
            addr,
            exc_info=True,
        )
    except Exception as e:
        logging.error("Error classifying job %s", args["file"], exc_info=True)
        clientsocket.sendall(
            json.dumps(
                {"error": f"Error classifying {traceback.format_exc()}"}
            ).encode()
        )
        raise e
    finally:
        try:
            clientsocket.close()
        except:
            pass


def read_all(socket):
    size = 4096
    data = bytearray()

    while size > 0:
        packet = socket.recv(size)
        data.extend(packet)
        if len(packet) < size:
            break
    return data


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


if __name__ == "__main__":
    main()
